# Author: Óscar Nájera
# License: 3-clause BSD
"""reST file generator.

Generate the rst files for the examples by iterating over the python
example files.

Files that generate images should start with 'plot'.
"""

import ast
import codeop
import contextlib
import copy
import gc
import importlib
import inspect
import os
import re
import stat
import sys
import traceback
import warnings
from functools import lru_cache
from io import StringIO
from pathlib import Path
from shutil import copyfile
from textwrap import indent
from time import time

import sphinx.util
from sphinx.errors import ConfigError, ExtensionError
from sphinx.util.console import blue, bold, red

from . import glr_path_static, py_source_parser
from .backreferences import (
    THUMBNAIL_PARENT_DIV,
    THUMBNAIL_PARENT_DIV_CLOSE,
    _make_ref_regex,
    _thumbnail_div,
    _write_backreferences,
    identify_names,
)
from .block_parser import BlockParser
from .interactive_example import (
    _add_jupyterlite_badge_logo,
    gen_binder_rst,
    gen_jupyterlite_rst,
)
from .notebook import jupyter_notebook, save_notebook
from .scrapers import (
    ImagePathIterator,
    _find_image_ext,
    _reset_dict,
    _scraper_dict,
    clean_modules,
    save_figures,
)
from .utils import (
    _W_KW,
    _collect_gallery_files,
    _combine_backreferences,
    _format_toctree,
    _replace_md5,
    _write_json,
    get_md5sum,
    optipng,
    scale_image,
    status_iterator,
    zip_files,
)

logger = sphinx.util.logging.getLogger("sphinx-gallery")


###############################################################################


class _LoggingTee:
    """A tee object to redirect streams to the logger."""

    def __init__(self, src_filename):
        self.logger = logger
        self.src_filename = src_filename
        self.logger_buffer = ""
        self.set_std_and_reset_position()

        # For TextIO compatibility
        self.closed = False
        self.encoding = "utf-8"

    def set_std_and_reset_position(self):
        if not isinstance(sys.stdout, _LoggingTee):
            self.origs = (sys.stdout, sys.stderr)
        sys.stdout = sys.stderr = self
        self.first_write = True
        self.output = StringIO()
        return self

    def restore_std(self):
        sys.stdout.flush()
        sys.stderr.flush()
        sys.stdout, sys.stderr = self.origs

    def write(self, data):
        self.output.write(data)

        if self.first_write:
            self.logger.verbose("Output from %s", self.src_filename, color="brown")
            self.first_write = False

        data = self.logger_buffer + data
        lines = data.splitlines()
        if data and data[-1] not in "\r\n":
            # Wait to write last line if it's incomplete. It will write next
            # time or when the LoggingTee is flushed.
            self.logger_buffer = lines[-1]
            lines = lines[:-1]
        else:
            self.logger_buffer = ""

        for line in lines:
            self.logger.verbose("%s", line)

    def flush(self):
        self.output.flush()
        if self.logger_buffer:
            self.logger.verbose("%s", self.logger_buffer)
            self.logger_buffer = ""

    # For TextIO compatibility
    def close(self):
        pass

    def fileno(self):
        return self.output.fileno()

    def isatty(self):
        return self.output.isatty()

    def readable(self):
        return False

    def seekable(self):
        return False

    def tell(self):
        return self.output.tell()

    def writable(self):
        return True

    @property
    def errors(self):
        return self.output.errors

    @property
    def newlines(self):
        return self.output.newlines

    # When called in gen_rst, conveniently use context managing
    def __enter__(self):
        return self

    def __exit__(self, type_, value, tb):
        self.restore_std()


###############################################################################
# The following strings are used when we have several pictures: we use
# an html div tag that our CSS uses to turn the lists into horizontal
# lists.
HLIST_HEADER = """
.. rst-class:: sphx-glr-horizontal

"""

HLIST_IMAGE_TEMPLATE = """
    *

      .. image:: /%s
            :class: sphx-glr-multi-img
"""

SINGLE_IMAGE = """
.. image:: /%s
    :class: sphx-glr-single-img
"""

CODE_OUTPUT = """.. rst-class:: sphx-glr-script-out

 .. code-block:: none

{0}\n"""

TIMING_CONTENT = """
.. rst-class:: sphx-glr-timing

   **Total running time of the script:** ({0:.0f} minutes {1:.3f} seconds)

"""

SPHX_GLR_SIG = """\n
.. only:: html

 .. rst-class:: sphx-glr-signature

    `Gallery generated by Sphinx-Gallery <https://sphinx-gallery.github.io>`_
"""

# Header used to include raw html
HTML_HEADER = """.. raw:: html

    <div class="output_subarea output_html rendered_html output_result">
{0}
    </div>
    <br />
    <br />"""

DOWNLOAD_LINKS_HEADER = """
.. _sphx_glr_download_{0}:

.. only:: html

  .. container:: sphx-glr-footer sphx-glr-footer-example
"""

CODE_DOWNLOAD = """
    .. container:: sphx-glr-download sphx-glr-download-python

      :download:`Download {1} source code: {0} <{0}>`
"""

NOTEBOOK_DOWNLOAD = """
    .. container:: sphx-glr-download sphx-glr-download-jupyter

      :download:`Download Jupyter notebook: {0} <{0}>`
"""

ZIP_DOWNLOAD = """
    .. container:: sphx-glr-download sphx-glr-download-zip

      :download:`Download zipped: {0} <{0}>`
"""

RECOMMENDATIONS_INCLUDE = """\n
.. include:: {0}.recommendations
"""


def codestr2rst(codestr, lang="python", lineno=None):
    """Return reStructuredText code block from code string."""
    if lineno is not None:
        # Sphinx only starts numbering from the first non-empty line.
        blank_lines = codestr.count("\n", 0, -len(codestr.lstrip()))
        lineno = f"   :lineno-start: {lineno + blank_lines}\n"
    else:
        lineno = ""
    # If the whole block is indented, prevent Sphinx from removing too much whitespace
    dedent = "   :dedent: 1\n"
    for line in codestr.splitlines():
        if line and not line.startswith((" ", "\t")):
            dedent = ""
            break
    code_directive = f".. code-block:: {lang}\n{dedent}{lineno}\n"
    indented_block = indent(codestr, " " * 4)
    return code_directive + indented_block


def _regroup(x):
    x = x.groups()
    return x[0] + x[1].split(".")[-1] + x[2]


def _sanitize_rst(string):
    """Use regex to remove at least some sphinx directives."""
    # :class:`a.b.c <thing here>`, :ref:`abc <thing here>` --> thing here
    p, e = r"(\s|^):[^:\s]+:`", r"`(\W|$)"
    string = re.sub(p + r"\S+\s*<([^>`]+)>" + e, r"\1\2\3", string)
    # :class:`~a.b.c` --> c
    string = re.sub(p + r"~([^`]+)" + e, _regroup, string)
    # :class:`a.b.c` --> a.b.c
    string = re.sub(p + r"([^`]+)" + e, r"\1\2\3", string)

    # ``whatever thing`` --> whatever thing
    p = r"(\s|^)`"
    string = re.sub(p + r"`([^`]+)`" + e, r"\1\2\3", string)
    # `~mymodule.MyClass` --> MyClass
    string = re.sub(p + r"~([^`]+)" + e, _regroup, string)

    # `whatever thing` --> whatever thing
    # `.MyClass` --> MyClass
    string = re.sub(p + r"\.?([^`]+)" + e, r"\1\2\3", string)

    # **string** --> string
    string = re.sub(r"\*\*([^\*]*)\*\*", r"\1", string)
    # *string* --> string
    string = re.sub(r"\*([^\*]*)\*", r"\1", string)
    # `link text <url>`_ --> link text
    string = re.sub(r"`([^`<>]+) <[^`<>]+>`\_\_?", r"\1", string)

    # :anchor:`the term` --> the term
    string = re.sub(r":[a-z]+:`([^`<>]+)( <[^`<>]+>)?`", r"\1", string)

    # r'\\dfrac' --> r'\dfrac'
    string = string.replace("\\\\", "\\")

    return string


def extract_intro_and_title(filename, docstring):
    """Extract and clean the first paragraph of module-level docstring."""
    # lstrip is just in case docstring has a '\n\n' at the beginning
    paragraphs = docstring.lstrip().split("\n\n")
    # remove comments and other syntax like `.. _link:`
    paragraphs = [p for p in paragraphs if not p.startswith(".. ") and len(p) > 0]
    if len(paragraphs) == 0:
        raise ExtensionError(
            "Example docstring should have a header for the example title. "
            "Please check the example file:\n {}\n".format(filename)
        )
    # Title is the first paragraph with any reStructuredText title chars
    # removed, i.e. lines that consist of (3 or more of the same) 7-bit
    # non-ASCII chars.
    # This conditional is not perfect but should hopefully be good enough.
    title_paragraph = paragraphs[0]
    match = re.search(r"^(?!([\W _])\1{3,})(.+)", title_paragraph, re.MULTILINE)

    if match is None:
        raise ExtensionError(
            f"Could not find a title in first paragraph:\n{title_paragraph}"
        )
    title = match.group(0).strip()
    # Use the title if no other paragraphs are provided
    intro_paragraph = title if len(paragraphs) < 2 else paragraphs[1]
    # Concatenate all lines of the first paragraph and truncate at 95 chars
    intro = re.sub("\n", " ", intro_paragraph)
    intro = _sanitize_rst(intro)
    title = _sanitize_rst(title)

    return intro, title


def md5sum_is_current(src_file, mode="b"):
    """Checks whether src_file has the same md5 hash as the one on disk."""
    src_md5 = get_md5sum(src_file, mode=mode)

    src_md5_file = str(src_file) + ".md5"
    if not os.path.exists(src_md5_file):
        return False

    with open(src_md5_file) as file_cs:
        ref_md5 = file_cs.read()

    return src_md5 == ref_md5


def save_thumbnail(image_path_template, src_file, script_vars, file_conf, gallery_conf):
    """Generate and Save the thumbnail image.

    Parameters
    ----------
    image_path_template : str
        holds the template where to save and how to name the image
    src_file : str
        path to source python file
    script_vars : dict
        Configuration and run time variables
    file_conf : dict
        File-specific settings given in source file comments as:
        ``# sphinx_gallery_<name> = <value>``
    gallery_conf : dict
        Sphinx-Gallery configuration dictionary
    """
    thumb_dir = os.path.join(os.path.dirname(image_path_template), "thumb")
    os.makedirs(thumb_dir, exist_ok=True)

    # read specification of the figure to display as thumbnail from main text
    thumbnail_number = file_conf.get("thumbnail_number", None)
    thumbnail_path = file_conf.get("thumbnail_path", None)
    # thumbnail_number has priority.
    if thumbnail_number is None and thumbnail_path is None:
        # If no number AND no path, set to default thumbnail_number
        thumbnail_number = 1
    if thumbnail_number is None:
        image_path = os.path.join(gallery_conf["src_dir"], thumbnail_path)
    else:
        if not isinstance(thumbnail_number, int):
            raise ExtensionError(
                "sphinx_gallery_thumbnail_number setting is not a number, "
                f"got {thumbnail_number!r}"
            )
        # negative index means counting from the last one
        if thumbnail_number < 0:
            thumbnail_number += len(script_vars["image_path_iterator"]) + 1
        image_path = image_path_template.format(thumbnail_number)
    del thumbnail_number, thumbnail_path, image_path_template
    thumbnail_image_path, ext = _find_image_ext(image_path)

    base_image_name = os.path.splitext(os.path.basename(src_file))[0]
    thumb_file = os.path.join(thumb_dir, f"sphx_glr_{base_image_name}_thumb.{ext}")

    if "formatted_exception" in script_vars and file_conf.get(
        "failing_thumbnail", True
    ):
        img = os.path.join(glr_path_static(), "broken_example.png")
    elif os.path.exists(thumbnail_image_path):
        img = thumbnail_image_path
    elif not os.path.exists(thumb_file):
        # create something to replace the thumbnail
        default_thumb_path = gallery_conf["default_thumb_file"]
        if default_thumb_path is None:
            default_thumb_path = os.path.join(
                glr_path_static(),
                "no_image.png",
            )
        img, ext = _find_image_ext(default_thumb_path)
    else:
        return
    # update extension, since gallery_conf setting can be different
    # from file_conf
    # Here we have to do .new.ext so that optipng and PIL behave well
    thumb_file = f"{os.path.splitext(thumb_file)[0]}.new.{ext}"
    if ext in ("svg", "gif"):
        copyfile(img, thumb_file)
    else:
        scale_image(img, thumb_file, *gallery_conf["thumbnail_size"])
        if "thumbnails" in gallery_conf["compress_images"]:
            optipng(thumb_file, gallery_conf["compress_images_args"])
    fname_old = f"{os.path.splitext(thumb_file)[0][:-3]}{ext}"
    _replace_md5(thumb_file, fname_old=fname_old)


def _get_gallery_header(dir_, gallery_conf, raise_error=True):
    """Get gallery header from GALLERY_HEADER.[ext] or README.[ext] file.

    Returns `None` if user supplied an index.rst or no gallery header file
    found and `raise_error=False`.
    """
    # First check if user supplies an index.rst and that index.rst is in the
    # copyfile regexp:
    if re.match(gallery_conf["copyfile_regex"], "index.rst"):
        fpth = os.path.join(dir_, "index.rst")
        if os.path.isfile(fpth):
            return None
    # Next look for GALLERY_HEADER.[ext] (and for backward-compatibility README.[ext]
    extensions = [".txt"] + sorted(gallery_conf["source_suffix"])
    for ext in extensions:
        for fname in ("GALLERY_HEADER", "README", "readme"):
            fpth = os.path.join(dir_, fname + ext)
            if os.path.isfile(fpth):
                return fpth
    if raise_error:
        raise ExtensionError(
            "Example directory {} does not have a GALLERY_HEADER file with "
            "one of the expected file extensions {}. Please write one to "
            "introduce your gallery.".format(dir_, extensions)
        )
    return None


def _write_subsection_index(
    gallery_conf,
    user_index_rst,
    is_subsection,
    target_dir,
    index_content,
    toctree_filenames,
):
    """Write `index.rst` file for subsection if user has not provided index file.

    Returns path to index file written or `None` if no index file written as user
    provided one.
    """
    index_path = None
    if gallery_conf["nested_sections"] and not user_index_rst and is_subsection:
        index_path = os.path.join(target_dir, "index.rst.new")
        head_ref = os.path.relpath(target_dir, gallery_conf["src_dir"])
        with open(index_path, "w", **_W_KW) as findex:
            findex.write(
                "\n\n.. _sphx_glr_{}:\n\n".format(head_ref.replace(os.sep, "_"))
            )
            findex.write(index_content)
            # Create toctree with all gallery examples and add to index file
            if len(toctree_filenames) > 0:
                subsection_index_toctree = _format_toctree(toctree_filenames)
                findex.write(subsection_index_toctree)

    return index_path


def _copy_non_example_files(gallery_conf, src_dir, header_fname, target_dir):
    """Copy non-example files to `target_dir`."""
    copyregex = gallery_conf["copyfile_regex"]
    if copyregex:
        listdir = [fname for fname in os.listdir(src_dir) if re.match(copyregex, fname)]
        if header_fname:
            # Don't copy over the gallery_header file
            listdir = [fname for fname in listdir if fname != Path(header_fname).name]
        for fname in listdir:
            src_file = os.path.normpath(os.path.join(src_dir, fname))
            target_file = os.path.join(target_dir, fname)
            _replace_md5(src_file, fname_old=target_file, method="copy")


def generate_dir_rst(
    src_dir,
    target_dir,
    gallery_conf,
    seen_backrefs,
    is_subsection=True,
):
    """Generate output example reST files for one gallery (sub)directory.

    Parameters
    ----------
    src_dir: str,
        Path to root or sub gallery directory containing example files
    target_dir: str,
        Path where parsed examples (rst, python files, etc) will be outputted
    gallery_conf : Dict[str, Any]
        Gallery configurations.
    seen_backrefs: set,
        Back references encountered when parsing this gallery
        will be stored in this set.
    is_subsection: bool,
        Weather `src_dir` is a subsection dir. If subsection dir, we write
        a `index.rst` file with toctree listing every example file.
        Default=True.

    Returns
    -------
    index_path: str or None
        Path to index rst file for the `src_dir`. None if user provided
        own index.
    index_content: str or None
        Gallery header content. `None` when user provided own index.rst.
    costs: List[Dict]
        List of dicts of costs for building each element of the gallery
         with keys "t", "mem", "src_file", and "target_dir".
    toctree_items: list,
        List of example file names we generated ReST for.
    backrefs_dir : dict[str, tuple]
        Dictionary where value is the backreference object and value
        is a tuple containing: example filename, full path to example source directory,
        full path to example target directory, intro, title.

    """
    index_content = ""
    # `_get_gallery_header` returns `None` if user supplied `index.rst`
    header_fname = _get_gallery_header(src_dir, gallery_conf)
    user_index_rst = True
    if header_fname:
        user_index_rst = False
        with open(header_fname, "r", encoding="utf-8") as fid:
            header_content = fid.read()
            index_content += header_content

    # Add empty lines to avoid bug in issue #165
    index_content += "\n\n"

    # Make all dirs ahead of time to avoid collisions in parallel processing
    os.makedirs(target_dir, exist_ok=True)
    image_dir = os.path.join(target_dir, "images")
    os.makedirs(image_dir, exist_ok=True)
    thumb_dir = os.path.join(image_dir, "thumb")
    os.makedirs(thumb_dir, exist_ok=True)
    if gallery_conf["jupyterlite"] is not None:
        _add_jupyterlite_badge_logo(image_dir)

    # Get example filenames from `src_dir`
    listdir = _collect_gallery_files([src_dir], gallery_conf)
    # sort them
    sorted_listdir = sorted(
        listdir, key=_get_callables(gallery_conf, "within_subsection_order", src_dir)[0]
    )

    # Add div containing all thumbnails;
    # this is helpful for controlling grid or flexbox behaviours
    index_content += THUMBNAIL_PARENT_DIV

    costs = []
    toctree_filenames = []
    build_target_dir = os.path.relpath(target_dir, gallery_conf["src_dir"])
    iterator = status_iterator(
        sorted_listdir,
        f"generating gallery for {build_target_dir}... ",
        length=len(sorted_listdir),
    )

    backrefs_dir = {}

    parallel = list
    p_fun = generate_file_rst
    if gallery_conf["parallel"]:
        from joblib import Parallel, delayed

        p_fun = delayed(generate_file_rst)
        parallel = Parallel(
            n_jobs=gallery_conf["parallel"],
            pre_dispatch="n_jobs",
            batch_size=1,
            backend="loky",
        )

    results = parallel(
        p_fun(fname, target_dir, src_dir, gallery_conf) for fname in iterator
    )
    for fi, (intro, title, (t, mem), out_vars) in enumerate(results):
        fname = sorted_listdir[fi]
        src_file = os.path.normpath(os.path.join(src_dir, fname))
        gallery_conf["titles"][src_file] = title
        # n.b. non-executable files have none of these three variables defined,
        # so the last conditional must be "elif" not just "else"
        if "formatted_exception" in out_vars:
            assert "passing" not in out_vars
            assert "stale" not in out_vars
            gallery_conf["failing_examples"][src_file] = out_vars["formatted_exception"]
        elif "passing" in out_vars:
            assert "stale" not in out_vars
            gallery_conf["passing_examples"].append(src_file)
        elif "stale" in out_vars:
            gallery_conf["stale_examples"].append(out_vars["stale"])
        costs.append(dict(t=t, mem=mem, src_file=src_file, target_dir=target_dir))
        gallery_item_filename = (
            (Path(build_target_dir) / fname).with_suffix("").as_posix()
        )
        this_entry = _thumbnail_div(
            target_dir, gallery_conf["src_dir"], fname, intro, title
        )
        index_content += this_entry
        toctree_filenames.append("/" + gallery_item_filename)

        # Write backreferences
        if "backrefs" in out_vars:
            backrefs_example = _write_backreferences(
                out_vars["backrefs"],
                seen_backrefs,
                gallery_conf,
                src_dir,
                target_dir,
                fname,
                intro,
                title,
            )
            _combine_backreferences(backrefs_dir, backrefs_example)

    # Close thumbnail parent div
    index_content += THUMBNAIL_PARENT_DIV_CLOSE

    # Write index file if required
    index_path = _write_subsection_index(
        gallery_conf,
        user_index_rst,
        is_subsection,
        target_dir,
        index_content,
        toctree_filenames,
    )

    if user_index_rst:
        # User has supplied index.rst, so blank out the content
        index_content = None

    # Copy over any other (non-gallery-example) files.
    _copy_non_example_files(gallery_conf, src_dir, header_fname, target_dir)

    return (
        index_path,
        index_content,
        costs,
        toctree_filenames,
        backrefs_dir,
    )


def handle_exception(exc_info, src_file, script_vars, gallery_conf):
    """Trim and format exception, maybe raise error, etc."""
    from .gen_gallery import _expected_failing_examples

    etype, exc, tb = exc_info
    stack = traceback.extract_tb(tb)
    # The full traceback will look something like:
    #
    #   File "/home/larsoner/python/sphinx-gallery/sphinx_gallery/gen_rst.py...
    #     mem_max, _ = gallery_conf['call_memory'](
    #   File "/home/larsoner/python/sphinx-gallery/sphinx_gallery/gen_galler...
    #     mem, out = memory_usage(func, max_usage=True, retval=True,
    #   File "/home/larsoner/.local/lib/python3.8/site-packages/memory_profi...
    #     returned = f(*args, **kw)
    #   File "/home/larsoner/python/sphinx-gallery/sphinx_gallery/gen_rst.py...
    #     exec(self.code, self.fake_main.__dict__)
    #   File "/home/larsoner/python/sphinx-gallery/sphinx_gallery/tests/tiny...
    #     raise RuntimeError('some error')
    # RuntimeError: some error
    #
    # But we should trim these to just the relevant trace at the user level,
    # so we inspect the traceback to find the start and stop points.
    start = 0
    stop = len(stack)
    root = os.path.dirname(__file__) + os.sep
    for ii, s in enumerate(stack, 1):
        # Trim our internal stack
        if s.name.startswith("_sg_call_memory"):
            start = max(ii, start)
        elif s.filename.startswith(root + "gen_rst.py"):
            # SyntaxError
            if s.name == "execute_code_block" and (
                "compile(" in s.line or "save_figures" in s.line
            ):
                start = max(ii, start)
            # Any other error
            elif s.name == "__call__":
                start = max(ii, start)
            # Our internal input() check
            elif s.name == "_check_input" and ii == len(stack):
                stop = ii - 1
    stack = stack[start:stop]

    formatted_exception = "Traceback (most recent call last):\n" + "".join(
        traceback.format_list(stack) + traceback.format_exception_only(etype, exc)
    )

    expected = src_file in _expected_failing_examples(gallery_conf)
    src_file_rel = os.path.relpath(src_file, gallery_conf["src_dir"])
    if expected:
        func, color, kind = logger.info, blue, "expectedly"
    else:
        func, color, kind = logger.warning, red, "unexpectedly"
    func(  # needs leading newline to get away from iterator
        f"\n{bold(color('%s'))} {kind} failed to execute correctly:\n\n%s",
        src_file_rel,
        color(indent(formatted_exception, "    ")),
    )

    except_rst = codestr2rst(formatted_exception, lang="pytb")

    # Breaks build on first example error
    if gallery_conf["abort_on_example_error"]:
        raise
    # Stores failing file
    script_vars["formatted_exception"] = formatted_exception
    script_vars["execute_script"] = False

    # Ensure it's marked as our style
    except_rst = ".. rst-class:: sphx-glr-script-out\n\n" + except_rst
    return except_rst


# Adapted from github.com/python/cpython/blob/3.7/Lib/warnings.py
def _showwarning(message, category, filename, lineno, file=None, line=None):
    if file is None:
        file = sys.stderr
        if file is None:
            # sys.stderr is None when run with pythonw.exe:
            # warnings get lost
            return
    text = warnings.formatwarning(message, category, filename, lineno, line)
    try:
        file.write(text)
    except OSError:
        # the file (probably stderr) is invalid - this warning gets lost.
        pass


@contextlib.contextmanager
def patch_warnings():
    """Patch warnings.showwarning to actually write out the warning."""
    # Sphinx or logging or someone is patching warnings, but we want to
    # capture them, so let's patch over their patch...
    orig_showwarning = warnings.showwarning
    try:
        warnings.showwarning = _showwarning
        yield
    finally:
        warnings.showwarning = orig_showwarning


class _exec_once:
    """Deal with memory_usage calling functions more than once (argh)."""

    def __init__(self, code, fake_main):
        self.code = code
        self.fake_main = fake_main
        self.run = False

    def __call__(self):
        if not self.run:
            self.run = True
            old_main = sys.modules.get("__main__", None)
            with patch_warnings():
                sys.modules["__main__"] = self.fake_main
                try:
                    exec(self.code, self.fake_main.__dict__)
                finally:
                    if old_main is not None:
                        sys.modules["__main__"] = old_main


def _get_memory_base():
    """Get the base amount of memory used by the current Python process."""
    from memory_profiler import memory_usage

    memory_base = memory_usage(max_usage=True)
    return memory_base


def _get_parser(fname, gallery_conf):
    """Get parser and language."""
    if fname.endswith(".py"):
        parser = py_source_parser
        language = "Python"
    else:
        parser = BlockParser(fname, gallery_conf)
        language = parser.language
    return parser, language


def _check_reset_logging_tee(src_file):
    # Helper to deal with our tests not necessarily calling execute_script
    # but rather execute_code_block directly
    if isinstance(sys.stdout, _LoggingTee):
        logging_tee = sys.stdout
    else:
        logging_tee = _LoggingTee(src_file)
    logging_tee.set_std_and_reset_position()
    return logging_tee


def _exec_and_get_memory(compiler, *, code_ast, gallery_conf, script_vars):
    """Execute ast, capturing output if last line expression and get max mem usage.

    Parameters
    ----------
    compiler : codeop.Compile
        Compiler to compile AST of code block.
    code_ast : ast.Module
        AST parsed code to execute.
    gallery_conf : Dict[str, Any]
        Gallery configurations.
    script_vars : Dict[str, Any]
        Configuration and runtime variables.

    Returns
    -------
    is_last_expr : bool
        Whether the last expression in `code_ast` is an ast.Expr.
    mem_max : float
        Max memory used during execution.
    """
    src_file = script_vars["src_file"]
    # capture output if last line is expression
    is_last_expr = False
    call_memory, _ = _get_call_memory_and_base(gallery_conf)
    if len(code_ast.body) and isinstance(code_ast.body[-1], ast.Expr):
        is_last_expr = True
        last_val = code_ast.body.pop().value
        # exec body minus last expression
        mem_body, _ = call_memory(
            _exec_once(compiler(code_ast, src_file, "exec"), script_vars["fake_main"])
        )
        # exec last expression, made into assignment
        body = [
            ast.Assign(targets=[ast.Name(id="___", ctx=ast.Store())], value=last_val)
        ]
        # `type_ignores` empty list deals with: https://bugs.python.org/issue3589
        last_val_ast = ast.Module(body=body, type_ignores=[])
        ast.fix_missing_locations(last_val_ast)
        mem_last, _ = call_memory(
            _exec_once(
                compiler(last_val_ast, src_file, "exec"), script_vars["fake_main"]
            )
        )
        mem_max = max(mem_body, mem_last)
    else:
        mem_max, _ = call_memory(
            _exec_once(compiler(code_ast, src_file, "exec"), script_vars["fake_main"])
        )
    return is_last_expr, mem_max


def _get_last_repr(capture_repr, ___):
    """Get repr of last expression, using first method in 'capture_repr' available."""
    for meth in capture_repr:
        try:
            last_repr = getattr(___, meth)()
            # for case when last statement is print()
            if last_repr is None or last_repr == "None":
                repr_meth = None
            else:
                repr_meth = meth
        except Exception:
            last_repr = None
            repr_meth = None
        else:
            if isinstance(last_repr, str):
                break
    return last_repr, repr_meth


def _get_code_output(
    is_last_expr, example_globals, gallery_conf, logging_tee, images_rst, capture_repr
):
    """Obtain standard output and html output in reST.

    Parameters
    ----------
    is_last_expr : bool
        Whether the last expression in executed code is an ast.Expr.
    example_globals: Dict[str, Any]
        Global variables for examples.
    logging_tee : _LoggingTee
        Logging tee.
    images_rst : str
        rst code to embed the images in the document.
    gallery_conf : Dict[str, Any]
        Gallery configurations.
    file_conf : Dict[str, Any]
        File-specific settings given in source file comments as:
        ``# sphinx_gallery_<name> = <value>``.

    Returns
    -------
    code_output : str
        reST of output of executed code block, including images and captured output.
    """
    last_repr = None
    repr_meth = None
    if is_last_expr:
        # capture the last repr variable
        ___ = example_globals["___"]
        ignore_repr = False
        if gallery_conf["ignore_repr_types"]:
            ignore_repr = re.search(gallery_conf["ignore_repr_types"], str(type(___)))
        if capture_repr != () and not ignore_repr:
            last_repr, repr_meth = _get_last_repr(capture_repr, ___)

    captured_std = logging_tee.output.getvalue().expandtabs()
    # normal string output
    if repr_meth in ["__repr__", "__str__"] and last_repr:
        captured_std = f"{captured_std}\n{last_repr}"
    if captured_std and not captured_std.isspace():
        captured_std = CODE_OUTPUT.format(indent(captured_std, " " * 4))
    else:
        captured_std = ""

    # Sanitize ANSI escape characters for reST output
    ansi_escape = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")
    captured_std = ansi_escape.sub("", captured_std)

    # give html output its own header
    if repr_meth == "_repr_html_":
        captured_html = HTML_HEADER.format(indent(last_repr, " " * 4))
    else:
        captured_html = ""

    code_output = f"\n{images_rst}\n\n{captured_std}\n{captured_html}\n\n"
    return code_output


def _reset_cwd_syspath(cwd, sys_path):
    """Reset cwd and sys.path."""
    os.chdir(cwd)
    sys.path = sys_path


def execute_code_block(
    compiler, block, example_globals, script_vars, gallery_conf, file_conf
):
    """Execute the code block of the example file.

    Parameters
    ----------
    compiler : codeop.Compile
        Compiler to compile AST of code block.
    block : sphinx_gallery.py_source_parser.Block
        The code block to be executed.
    example_globals: Dict[str, Any]
        Global variables for examples.
    script_vars : Dict[str, Any]
        Configuration and runtime variables.
    gallery_conf : Dict[str, Any]
        Gallery configurations.
    file_conf : Dict[str, Any]
        File-specific settings given in source file comments as:
        ``# sphinx_gallery_<name> = <value>``.

    Returns
    -------
    code_output : str
        Output of executing code in reST.
    """
    if example_globals is None:  # testing shortcut
        example_globals = script_vars["fake_main"].__dict__
    # If example is not suitable to run, skip executing its blocks
    if not script_vars["execute_script"] or block.type == "text":
        return ""

    cwd = os.getcwd()
    # Redirect output to stdout
    src_file = script_vars["src_file"]
    logging_tee = _check_reset_logging_tee(src_file)
    assert isinstance(logging_tee, _LoggingTee)

    # First cd in the original example dir, so that any file
    # created by the example get created in this directory
    os.chdir(os.path.dirname(src_file))

    sys_path = copy.deepcopy(sys.path)
    sys.path.append(os.getcwd())

    # Save figures unless there is a `sphinx_gallery_defer_figures` flag
    defer_figs_match = re.search(
        r"^[\ \t]*#\s*sphinx_gallery_defer_figures[\ \t]*\n?",
        block.content,
        re.MULTILINE,
    )
    need_save_figures = defer_figs_match is None

    block_conf = py_source_parser.extract_file_config(block.content)

    # Add `sphinx_gallery_multi_image_block` setting to block variables
    # (extract config rather than just regex search since the option's value is needed)
    script_vars["multi_image"] = block_conf.get("multi_image_block")

    # Determine the block's `capture_repr` value by prioritising the block-level
    # `sphinx_gallery_capture_repr_block` setting, then the file-level
    # `sphinx_gallery_capture_repr` setting, and finally the
    # global `capture_repr` gallery setting.
    capture_repr = block_conf.get(
        "capture_repr_block",
        file_conf.get("capture_repr", gallery_conf["capture_repr"]),
    )

    # Add file_conf to script_vars to be read by image scrapers
    script_vars["file_conf"] = file_conf

    try:
        # The "compile" step itself can fail on a SyntaxError, so just prepend
        # newlines to get the correct failing line to show up in the traceback
        code_ast = compile(
            "\n" * (block.lineno - 1) + block.content,
            src_file,
            "exec",
            flags=ast.PyCF_ONLY_AST | compiler.flags,
            dont_inherit=1,
        )
        is_last_expr, mem_max = _exec_and_get_memory(
            compiler,
            code_ast=code_ast,
            gallery_conf=gallery_conf,
            script_vars=script_vars,
        )
        script_vars["memory_delta"].append(mem_max)
        # This should be inside the try block, e.g., in case of a savefig error
        logging_tee.restore_std()
        if need_save_figures:
            need_save_figures = False
            images_rst = save_figures(block, script_vars, gallery_conf)
        else:
            images_rst = ""
    except Exception:
        logging_tee.restore_std()
        except_rst = handle_exception(
            sys.exc_info(), src_file, script_vars, gallery_conf
        )
        code_output = f"\n{except_rst}\n\n\n\n"
        # still call this even though we won't use the images so that
        # figures are closed
        if need_save_figures:
            save_figures(block, script_vars, gallery_conf)
    else:
        _reset_cwd_syspath(cwd, sys_path)

        code_output = _get_code_output(
            is_last_expr,
            example_globals,
            gallery_conf,
            logging_tee,
            images_rst,
            capture_repr,
        )
    finally:
        _reset_cwd_syspath(cwd, sys_path)
        logging_tee.restore_std()

    return code_output


def executable_script(src_file, gallery_conf):
    """Validate if script has to be run according to gallery configuration.

    Parameters
    ----------
    src_file : str
        path to python script

    gallery_conf : dict
        Contains the configuration of Sphinx-Gallery

    Returns
    -------
    bool
        True if script has to be executed
    """
    filename_pattern = gallery_conf["filename_pattern"]
    execute = re.search(filename_pattern, src_file) and gallery_conf["plot_gallery"]
    return execute


def _check_input(prompt=None):
    raise ExtensionError(
        "Cannot use input() builtin function in Sphinx-Gallery examples"
    )


def execute_script(script_blocks, script_vars, gallery_conf, file_conf):
    """Execute and capture output from python script already in block structure.

    Parameters
    ----------
    script_blocks : list
        (label, content, line_number)
        List where each element is a tuple with the label ('text' or 'code'),
        the corresponding content string of block and the leading line number
    script_vars : dict
        Configuration and run time variables
    gallery_conf : dict
        Contains the configuration of Sphinx-Gallery
    file_conf : dict
        File-specific settings given in source file comments as:
        ``# sphinx_gallery_<name> = <value>``

    Returns
    -------
    output_blocks : list
        List of strings where each element is the restructured text
        representation of the output of each block
    time_elapsed : float
        Time elapsed during execution
    """
    # Examples may contain if __name__ == '__main__' guards
    # for in example scikit-learn if the example uses multiprocessing.
    # Here we create a new __main__ module, and temporarily change
    # sys.modules when running our example
    call_memory, _ = _get_call_memory_and_base(gallery_conf)

    fake_main = importlib.util.module_from_spec(
        importlib.util.spec_from_loader("__main__", None)
    )
    example_globals = fake_main.__dict__

    example_globals.update(
        {
            # A lot of examples contains 'print(__doc__)' for example in
            # scikit-learn so that running the example prints some useful
            # information. Because the docstring has been separated from
            # the code blocks in sphinx-gallery, __doc__ is actually
            # __builtin__.__doc__ in the execution context and we do not
            # want to print it
            "__doc__": "",
            # Don't ever support __file__: Issues #166 #212
            # Don't let them use input()
            "input": _check_input,
        }
    )
    script_vars["example_globals"] = example_globals

    argv_orig = sys.argv[:]
    if script_vars["execute_script"]:
        # We want to run the example without arguments. See
        # https://github.com/sphinx-gallery/sphinx-gallery/pull/252
        # for more details.
        sys.argv[0] = script_vars["src_file"]
        (reset_argv,) = _get_callables(gallery_conf, "reset_argv")
        sys.argv[1:] = reset_argv(gallery_conf, script_vars)
        gc.collect()
        memory_start, _ = call_memory(lambda: None)
    else:
        memory_start = 0.0

    t_start = time()
    compiler = codeop.Compile()
    # include at least one entry to avoid max() ever failing
    script_vars["memory_delta"] = [memory_start]
    script_vars["fake_main"] = fake_main
    output_blocks = list()
    with _LoggingTee(script_vars.get("src_file", "")) as logging_tee:
        for block in script_blocks:
            logging_tee.set_std_and_reset_position()
            output_blocks.append(
                execute_code_block(
                    compiler,
                    block,
                    example_globals,
                    script_vars,
                    gallery_conf,
                    file_conf,
                )
            )
    time_elapsed = time() - t_start
    sys.argv = argv_orig
    script_vars["memory_delta"] = max(script_vars["memory_delta"])
    if script_vars["execute_script"]:
        script_vars["memory_delta"] -= memory_start
        # Write md5 checksum if the example was meant to run (no-plot
        # shall not cache md5sum) and has built correctly
        with open(script_vars["target_file"] + ".md5", "w") as file_cs:
            file_cs.write(get_md5sum(script_vars["target_file"], mode="t"))
        script_vars["passing"] = True

    return output_blocks, time_elapsed


def _make_dummy_images(executable, file_conf, script_vars):
    """Make dummy images when not executing the example."""
    if not executable:
        dummy_image = file_conf.get("dummy_images", None)
        if dummy_image is not None:
            if isinstance(dummy_image, bool) or not isinstance(dummy_image, int):
                raise ExtensionError(
                    "sphinx_gallery_dummy_images setting is not an integer, "
                    "got {dummy_image!r}"
                )

            image_path_iterator = script_vars["image_path_iterator"]
            stock_img = os.path.join(glr_path_static(), "no_image.png")
            for _, path in zip(range(dummy_image), image_path_iterator):
                if not os.path.isfile(path):
                    copyfile(stock_img, path)


def _clean_script_blocks(gallery_conf, parser, script_blocks, output_blocks):
    """Remove ignore blocks, config comments and final empty blocks."""
    # Ignore blocks must be processed before the
    # remaining config comments are removed.
    script_blocks = [
        py_source_parser.Block(label, parser.remove_ignore_blocks(content), line_number)
        for label, content, line_number in script_blocks
    ]

    if gallery_conf["remove_config_comments"]:
        script_blocks = [
            py_source_parser.Block(
                label, parser.remove_config_comments(content), line_number
            )
            for label, content, line_number in script_blocks
        ]

    # Remove final empty block, (can occur after config comments removed)
    if script_blocks[-1].content.isspace():
        script_blocks = script_blocks[:-1]
        output_blocks = output_blocks[:-1]

    return script_blocks, output_blocks


def _get_backreferences(gallery_conf, script_vars, script_blocks, node, target_file):
    """Get example backreferences for `script_blocks` and write _codeobj."""
    if gallery_conf["inspect_global_variables"]:
        global_variables = script_vars["example_globals"]
    else:
        global_variables = None
    ref_regex = _make_ref_regex(gallery_conf["default_role"])
    example_code_obj = identify_names(script_blocks, ref_regex, global_variables, node)
    if example_code_obj:
        _write_json(target_file, example_code_obj, ".codeobj")
    exclude_regex = gallery_conf["exclude_implicit_doc_regex"]

    def _normalize_name(cobj):
        full_name = "{module}.{name}".format(**cobj)
        for pattern in gallery_conf["prefer_full_module"]:
            if re.search(pattern, full_name):
                return full_name
        return "{module_short}.{name}".format(**cobj)

    backrefs = {
        _normalize_name(cobj)
        for cobjs in example_code_obj.values()
        for cobj in cobjs
        if cobj["module"].startswith(gallery_conf["doc_module"])
        and (
            cobj["is_explicit"]
            or (not exclude_regex)
            or (not exclude_regex.search("{module}.{name}".format(**cobj)))
        )
    }

    # This can help with garbage collection in some instances
    if global_variables is not None and "___" in global_variables:
        del global_variables["___"]
    del global_variables

    return backrefs


def generate_file_rst(fname, target_dir, src_dir, gallery_conf):
    """Generate the rst file for a given example.

    Parameters
    ----------
    fname : str
        Filename of python script.
    target_dir : str
        Absolute path to directory in documentation where examples are saved.
    src_dir : str
        Absolute path to directory where source examples are stored.
    gallery_conf : dict
        Contains the configuration of Sphinx-Gallery.

    Returns
    -------
    intro: str
        The introduction of the example.
    title : str
        The example title.
    cost : tuple
        A tuple containing the ``(time_elapsed, memory_used)`` required to run the
        script.
    out_vars : dict
        Variables used to run the script, possibly with entries:

        "stale"
            True if the example was stale.
        "backrefs"
            The backreferences.
        "passing"
            True if the example passed.
        "formatted_exception"
            Formatted string of the exception.
    """
    src_file = os.path.normpath(os.path.join(src_dir, fname))
    out_vars = dict()
    target_file = Path(target_dir) / fname
    _replace_md5(src_file, target_file, method="copy", mode="t")

    parser, language = _get_parser(fname, gallery_conf)

    file_conf, script_blocks, node = parser.split_code_and_text_blocks(
        src_file, return_node=True
    )

    intro, title = extract_intro_and_title(fname, script_blocks[0].content)

    executable = executable_script(src_file, gallery_conf)

    if md5sum_is_current(target_file, mode="t"):
        do_return = True
        logger.debug(f"md5sum is current: {target_file}")
        if executable:
            if gallery_conf["run_stale_examples"]:
                do_return = False
            else:
                out_vars["stale"] = str(target_file)
        if do_return:
            return intro, title, (0, 0), out_vars

    image_dir = os.path.join(target_dir, "images")
    os.makedirs(image_dir, exist_ok=True)

    base_image_name = os.path.splitext(fname)[0]
    image_fname = "sphx_glr_" + base_image_name + "_{0:03}.png"
    image_path_template = os.path.join(image_dir, image_fname)

    script_vars = {
        "execute_script": executable,
        "image_path_iterator": ImagePathIterator(image_path_template),
        "src_file": src_file,
        "target_file": str(target_file),
    }

    if executable and gallery_conf["reset_modules_order"] in ["before", "both"]:
        clean_modules(gallery_conf, fname, "before")
    output_blocks, time_elapsed = execute_script(
        script_blocks, script_vars, gallery_conf, file_conf
    )

    logger.debug("%s ran in : %.2g seconds", src_file, time_elapsed)

    # Create dummy images
    _make_dummy_images(executable, file_conf, script_vars)

    script_blocks, output_blocks = _clean_script_blocks(
        gallery_conf,
        parser,
        script_blocks,
        output_blocks,
    )

    example_rst = rst_blocks(
        script_blocks,
        output_blocks,
        file_conf,
        gallery_conf,
        language=language,
    )
    _, memory_base = _get_call_memory_and_base(gallery_conf)
    memory_used = memory_base + script_vars["memory_delta"]
    if not executable:
        time_elapsed = memory_used = 0.0  # don't let the output change
    save_rst_example(
        example_rst,
        target_file,
        time_elapsed,
        memory_used,
        gallery_conf,
        language=language,
    )

    save_thumbnail(image_path_template, src_file, script_vars, file_conf, gallery_conf)
    files_to_zip = [target_file]

    if target_file.suffix in gallery_conf["notebook_extensions"]:
        example_nb = jupyter_notebook(script_blocks, gallery_conf, target_dir)
        ipy_fname = target_file.with_suffix(".ipynb.new")
        save_notebook(example_nb, ipy_fname)
        _replace_md5(ipy_fname, mode="t")
        files_to_zip += [target_file.with_suffix(".ipynb")]

    # Produce the zip file of all sources
    zip_files(files_to_zip, target_file.with_suffix(".zip"), target_dir)

    # Get names
    out_vars["backrefs"] = _get_backreferences(
        gallery_conf,
        script_vars,
        script_blocks,
        node,
        target_file,
    )
    for key in ("passing", "formatted_exception"):
        if key in script_vars:
            out_vars[key] = script_vars[key]
    # don't keep this during reset
    del script_vars
    if executable and gallery_conf["reset_modules_order"] in ["after", "both"]:
        clean_modules(gallery_conf, fname, "after")

    return intro, title, (time_elapsed, memory_used), out_vars


EXAMPLE_HEADER = """
.. DO NOT EDIT.
.. THIS FILE WAS AUTOMATICALLY GENERATED BY SPHINX-GALLERY.
.. TO MAKE CHANGES, EDIT THE SOURCE PYTHON FILE:
.. "{0}"
.. LINE NUMBERS ARE GIVEN BELOW.

.. only:: html

    .. note::
        :class: sphx-glr-download-link-note

        :ref:`Go to the end <sphx_glr_download_{1}>`
        to download the full example code.{2}

.. rst-class:: sphx-glr-example-title

.. _sphx_glr_{1}:

"""
RST_BLOCK_HEADER = """\
.. GENERATED FROM PYTHON SOURCE LINES {0}-{1}

"""


def rst_blocks(
    script_blocks, output_blocks, file_conf, gallery_conf, *, language="python"
):
    """Generate the rst string containing the script prose, code and output.

    Parameters
    ----------
    script_blocks : list
        (label, content, line_number)
        List where each element is a tuple with the label ('text' or 'code'),
        the corresponding content string of block and the leading line number
    output_blocks : list
        List of strings where each element is the restructured text
        representation of the output of each block
    file_conf : dict
        File-specific settings given in source file comments as:
        ``# sphinx_gallery_<name> = <value>``
    language : str
        The language to be used for syntax highlighting in code blocks. Must be a name
        or alias recognized by Pygments.
    gallery_conf : dict
        Contains the configuration of Sphinx-Gallery

    Returns
    -------
    out : str
        rst notebook
    """
    # A simple example has two blocks: one for the
    # example introduction/explanation and one for the code
    is_example_notebook_like = len(script_blocks) > 2
    example_rst = ""
    for bi, (script_block, code_output) in enumerate(zip(script_blocks, output_blocks)):
        # do not add comment to the title block, otherwise the linking does
        # not work properly
        if bi > 0:
            example_rst += RST_BLOCK_HEADER.format(
                script_block.lineno,
                script_block.lineno + script_block.content.count("\n"),
            )
        if script_block.type == "code":
            lineno = (
                script_block.lineno
                if file_conf.get("line_numbers", gallery_conf["line_numbers"])
                else None
            )

            code_rst = (
                codestr2rst(script_block.content, lang=language, lineno=lineno) + "\n"
            )
            if is_example_notebook_like:
                example_rst += code_rst
                example_rst += code_output
            else:
                example_rst += code_output
                if "sphx-glr-script-out" in code_output:
                    # Add some vertical space after output
                    example_rst += "\n\n|\n\n"
                example_rst += code_rst
        else:
            block_separator = (
                "\n\n" if not script_block.content.endswith("\n") else "\n"
            )
            example_rst += script_block.content + block_separator

    return example_rst


def save_rst_example(
    example_rst,
    example_file,
    time_elapsed,
    memory_used,
    gallery_conf,
    *,
    language="python",
):
    """Saves the rst notebook to example_file including header & footer.

    Parameters
    ----------
    example_rst : str
        rst containing the executed file content
    example_file : str
        Filename with full path of python example file in documentation folder
    language : str
        Name of the programming language the example is in
    time_elapsed : float
        Time elapsed in seconds while executing file
    memory_used : float
        Additional memory used during the run.
    gallery_conf : dict
        Sphinx-Gallery configuration dictionary
    """
    example_file = Path(example_file)
    example_fname = str(example_file.relative_to(gallery_conf["src_dir"]))
    ref_fname = example_fname.replace(os.sep, "_")

    binder_conf = gallery_conf["binder"]
    is_binder_enabled = len(binder_conf) > 0

    jupyterlite_conf = gallery_conf["jupyterlite"]
    is_jupyterlite_enabled = jupyterlite_conf is not None

    interactive_example_text = ""
    if is_binder_enabled or is_jupyterlite_enabled:
        interactive_example_text += " or to run this example in your browser via "

    if is_binder_enabled and is_jupyterlite_enabled:
        interactive_example_text += "JupyterLite or Binder"
    elif is_binder_enabled:
        interactive_example_text += "Binder"
    elif is_jupyterlite_enabled:
        interactive_example_text += "JupyterLite"

    example_rst = (
        EXAMPLE_HEADER.format(example_fname, ref_fname, interactive_example_text)
        + example_rst
    )

    if time_elapsed > gallery_conf["min_reported_time"]:
        time_m, time_s = divmod(time_elapsed, 60)
        example_rst += TIMING_CONTENT.format(time_m, time_s)

    if gallery_conf["show_memory"]:
        example_rst += f"**Estimated memory usage:** {memory_used: .0f} MB\n\n"

    example_rst += DOWNLOAD_LINKS_HEADER.format(ref_fname)

    save_notebook = example_file.suffix in gallery_conf["notebook_extensions"]

    # Generate a binder URL if specified
    if is_binder_enabled and save_notebook:
        binder_badge_rst = gen_binder_rst(example_file, binder_conf, gallery_conf)
        binder_badge_rst = indent(binder_badge_rst, "  ")  # need an extra two
        example_rst += binder_badge_rst

    if is_jupyterlite_enabled and save_notebook:
        jupyterlite_rst = gen_jupyterlite_rst(example_file, gallery_conf)
        jupyterlite_rst = indent(jupyterlite_rst, "  ")  # need an extra two
        example_rst += jupyterlite_rst

    if save_notebook:
        ipynb_download_file = example_file.with_suffix(".ipynb").name
        example_rst += NOTEBOOK_DOWNLOAD.format(ipynb_download_file)

    example_rst += CODE_DOWNLOAD.format(example_file.name, language)
    example_rst += ZIP_DOWNLOAD.format(example_file.with_suffix(".zip").name)

    if gallery_conf["recommender"]["enable"]:
        # extract the filename without the extension
        recommend_fname = Path(example_fname).stem
        example_rst += RECOMMENDATIONS_INCLUDE.format(recommend_fname)

    if gallery_conf["show_signature"]:
        example_rst += SPHX_GLR_SIG

    write_file_new = example_file.with_suffix(".rst.new")
    with open(write_file_new, "w", **_W_KW) as f:
        f.write(example_rst)
    # make it read-only so that people don't try to edit it
    mode = os.stat(write_file_new).st_mode
    ro_mask = 0x777 ^ (stat.S_IWRITE | stat.S_IWGRP | stat.S_IWOTH)
    os.chmod(write_file_new, mode & ro_mask)
    # in case it wasn't in our pattern, only replace the file if it's
    # still stale.
    _replace_md5(write_file_new, mode="t")


def _get_callables(gallery_conf, key, src_dir=None):
    """Get callables for the given conf key, returning tuple of callable(s).

    If value is a string, import, with the following exceptions:

    * `within_subsection_order` - add full path if value is a built-in aliases,
      instantiate if value is a class
    * `image_scrapers` - get value from `_scraper_dict` or import module and
      get `_get_sg_image_scraper` attribute
    * `reset_modules` - get value from `_reset_dict`
    """
    builtin_aliases = (
        "ExampleTitleSortKey",
        "FileNameSortKey",
        "FileSizeSortKey",
        "NumberOfCodeLinesSortKey",
    )
    singletons = (
        "reset_argv",
        "minigallery_sort_order",
        "subsection_order",
        "within_subsection_order",
    )
    # the following should be the case (internal use only):
    assert key in ("image_scrapers", "reset_modules", "jupyterlite") + singletons, key
    # Get correct config value and ensure it's a list
    which = gallery_conf[key]
    if key == "jupyterlite":
        which = [which["notebook_modification_function"]]
    elif key in singletons:
        which = [which]
    if not isinstance(which, (tuple, list)):
        which = [which]
    which = list(which)

    for wi, what in enumerate(which):
        # Get readable string, for error message
        if key == "jupyterlite":
            readable = f"{key}['notebook_modification_function']"
        elif key in singletons:
            readable = f"{key}={repr(what)}"
        else:
            readable = f"{key}[{wi}]={repr(what)}"
        # Import string FQNs or get value from dict
        if isinstance(what, str):
            # use fully qualified name to resolve builtin callable classes
            # (otherwise not serializable)
            if what in builtin_aliases:
                what = f"sphinx_gallery.sorting.{what}"
            if "." in what:
                mod, attr = what.rsplit(".", 1)
                try:
                    what = getattr(importlib.import_module(mod), attr)
                except Exception:
                    raise ConfigError(
                        f"Unknown string option {readable} "
                        f"when importing {attr} from {mod}"
                    )
            elif key == "image_scrapers":
                if what in _scraper_dict:
                    what = _scraper_dict[what]
                else:
                    try:
                        what = importlib.import_module(what)
                        what = getattr(what, "_get_sg_image_scraper")
                        what = what()
                    except Exception:
                        raise ConfigError(f"Unknown string option for {readable}")
            elif key == "reset_modules":
                if what not in _reset_dict:
                    raise ConfigError(f"Unknown string option for {readable}: {what}")
                what = _reset_dict[what]
            which[wi] = what
        # make sure classes get instantiated (so they become callable)
        if key == "within_subsection_order" and inspect.isclass(what):
            what = what(src_dir)
            which[wi] = what
        if inspect.isclass(what):
            raise ConfigError(
                f"Got class rather than callable instance for {readable}: {what}"
            )
        if not callable(what):
            raise ConfigError(f"{readable} must be callable")
    return tuple(which)


# Default no-op version
def _sg_call_memory_noop(func):
    return 0.0, func()


def _get_call_memory_and_base(gallery_conf, *, update=False):
    show_memory = gallery_conf["show_memory"]

    # Default to no-op version
    call_memory = _sg_call_memory_noop
    memory_base = 0.0

    if show_memory and gallery_conf["plot_gallery"]:
        if callable(show_memory):
            call_memory = show_memory
        elif gallery_conf["parallel"]:
            if update:
                logger.warning(
                    f"{gallery_conf['show_memory']=} disabled due to "
                    f"{gallery_conf['parallel']=}."
                )
                gallery_conf["show_memory"] = False
        else:
            out = _get_memprof_call_memory()
            if out is not None:
                call_memory, memory_base = out
            elif update:
                gallery_conf["show_memory"] = False

    assert callable(call_memory)

    return call_memory, memory_base


def _sg_call_memory_memprof(func):
    from memory_profiler import memory_usage  # noqa

    mem, out = memory_usage(func, max_usage=True, retval=True, multiprocess=True)
    try:
        mem = mem[0]  # old MP always returned a list
    except TypeError:  # 'float' object is not subscriptable
        pass
    return mem, out


@lru_cache()
def _get_memprof_call_memory():
    try:
        from memory_profiler import memory_usage  # noqa
    except ImportError:
        logger.warning(
            "Please install 'memory_profiler' to enable peak memory measurements."
        )
        return None
    else:
        return _sg_call_memory_memprof, _get_memory_base()



"""
Sphinx Gallery
==============

"""

import os

# dev versions should have "dev" in them, stable should not.
# doc/conf.py makes use of this to set the version drop-down.
__version__ = "0.19.0"


def glr_path_static():
    """Returns path to packaged static files"""
    return os.path.abspath(os.path.join(os.path.dirname(__file__), "_static"))



# Author: Óscar Nájera
# License: 3-clause BSD
"""Link resolver objects."""

import gzip
import json
import os
import posixpath
import re
import shelve
import urllib.parse as urllib_parse
import urllib.request as urllib_request
from io import BytesIO
from pathlib import Path
from urllib.error import HTTPError, URLError

import sphinx.util
from sphinx.errors import ExtensionError
from sphinx.search import js_index

from .utils import _W_KW, _read_json, status_iterator

logger = sphinx.util.logging.getLogger("sphinx-gallery")


def _get_data(url):
    """Get data over http(s) or from a local file."""
    if urllib_parse.urlparse(url).scheme in ("http", "https"):
        user_agent = "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.11 (KHTML, like Gecko) Chrome/23.0.1271.64 Safari/537.11"  # noqa: E501
        headers = {"User-Agent": user_agent}
        req = urllib_request.Request(url, None, headers)
        resp = urllib_request.urlopen(req)
        encoding = resp.headers.get("content-encoding", "plain")
        data = resp.read()
        if encoding == "gzip":
            data = gzip.GzipFile(fileobj=BytesIO(data)).read()
        elif encoding != "plain":
            raise ExtensionError(f"unknown encoding {encoding!r}")
        data = data.decode("utf-8")
    else:
        with open(url, mode="r", encoding="utf-8") as fid:
            data = fid.read()

    return data


def get_data(url, gallery_dir):
    """Persistent dictionary usage to retrieve the search indexes."""
    cached_file = os.path.join(gallery_dir, "searchindex")
    search_index = shelve.open(cached_file)
    if url in search_index:
        data = search_index[url]
    else:
        data = _get_data(url)
        search_index[url] = data
    search_index.close()

    return data


def parse_sphinx_docopts(index):
    """Parse the Sphinx index for documentation options.

    Parameters
    ----------
    index : str
        The Sphinx index page

    Returns
    -------
    docopts : dict
        The documentation options from the page.
    """
    pos = index.find("var DOCUMENTATION_OPTIONS")
    if pos < 0:
        pos = index.find("const DOCUMENTATION_OPTIONS")  # Sphinx 7.2+
    if pos < 0:
        raise ExtensionError("Documentation options could not be found in index.")
    pos = index.find("{", pos)
    if pos < 0:
        raise ExtensionError("Documentation options could not be found in index.")
    endpos = index.find("};", pos)
    if endpos < 0:
        raise ExtensionError("Documentation options could not be found in index.")
    block = index[pos + 1 : endpos].strip()
    docopts = {}
    for line in block.splitlines():
        key, value = line.split(":", 1)
        key = key.strip().strip('"')

        value = value.strip()
        if value[-1] == ",":
            value = value[:-1].rstrip()
        if value[0] in "\"'":
            value = value[1:-1]
        elif value == "false":
            value = False
        elif value == "true":
            value = True
        else:
            try:
                value = int(value)
            except ValueError:
                # In Sphinx 1.7.5, URL_ROOT is a JavaScript fragment.
                # Ignoring this entry since URL_ROOT is not used
                # elsewhere.
                # https://github.com/sphinx-gallery/sphinx-gallery/issues/382
                continue

        docopts[key] = value

    return docopts


class SphinxDocLinkResolver:
    """Resolve documentation links using searchindex.js generated by Sphinx.

    Parameters
    ----------
    doc_url : str
        The base URL of the project website.
    relative : bool
        Return relative links (only useful for links to documentation of this
        package).
    """

    def __init__(self, config, doc_url, gallery_dir, relative=False):
        self.config = config
        self.doc_url = doc_url
        self.gallery_dir = gallery_dir
        self.relative = relative
        self._link_cache = {}

        if isinstance(doc_url, Path):
            index_url = os.path.join(doc_url, "index.html")
            searchindex_url = os.path.join(doc_url, "searchindex.js")
            docopts_url = os.path.join(doc_url, "_static", "documentation_options.js")
        else:
            if relative:
                raise ExtensionError(
                    "Relative links are only supported for local "
                    "URLs (doc_url cannot be absolute)"
                )
            index_url = doc_url + "/"
            searchindex_url = doc_url + "/searchindex.js"
            docopts_url = doc_url + "/_static/documentation_options.js"

        # detect if we are using relative links on a Windows system
        if os.name.lower() == "nt" and isinstance(doc_url, Path):
            if not relative:
                raise ExtensionError(
                    "You have to use relative=True for the local"
                    " package on a Windows system."
                )
            self._is_windows = True
        else:
            self._is_windows = False

        # Download and find documentation options. As of Sphinx 1.7, these
        # options are now kept in a standalone file called
        # 'documentation_options.js'. Since SphinxDocLinkResolver can be called
        # not only for the documentation which is being built but also ones
        # that are being referenced, we need to try and get the index page
        # first and if that doesn't work, check for the
        # documentation_options.js file.
        index = get_data(index_url, gallery_dir)
        if "var DOCUMENTATION_OPTIONS" in index:
            self._docopts = parse_sphinx_docopts(index)
        else:
            docopts = get_data(docopts_url, gallery_dir)
            self._docopts = parse_sphinx_docopts(docopts)

        # download and initialize the search index
        sindex = get_data(searchindex_url, gallery_dir)
        self._searchindex = js_index.loads(sindex)

    def _get_index_match(self, first, second):
        try:
            match = self._searchindex["objects"][first]
        except KeyError:
            return None
        else:
            if isinstance(match, dict):
                try:
                    match = match[second]
                except KeyError:
                    return None
            elif isinstance(match, (list, tuple)):  # Sphinx 5.0.0 dev
                try:
                    for item in match:
                        if item[4] == second:
                            match = item[:4]
                            break
                    else:
                        return None
                except Exception:
                    return None
        return match

    def _get_link_type(self, cobj, use_full_module=False):
        """Get a valid link and type_, False if not found."""
        module_type = "module" if use_full_module else "module_short"
        first, second = cobj[module_type], cobj["name"]
        match = self._get_index_match(first, second)
        if match is None and "." in second:  # possible class attribute
            first, second = second.split(".", 1)
            first = ".".join([cobj[module_type], first])
            match = self._get_index_match(first, second)
        if match is None:
            link = type_ = None
        else:
            fname_idx = match[0]
            objname_idx = str(match[1])
            anchor = match[3]
            type_ = self._searchindex["objtypes"][objname_idx]

            fname = self._searchindex["filenames"][fname_idx]
            # In 1.5+ Sphinx seems to have changed from .rst.html to only
            # .html extension in converted files. Find this from the options.
            ext = self._docopts.get("FILE_SUFFIX", ".rst.html")
            fname = os.path.splitext(fname)[0] + ext
            if self._is_windows:
                fname = fname.replace("/", "\\")
                link = os.path.join(self.doc_url, fname)
            else:
                link = posixpath.join(self.doc_url, fname)

            fullname = ".".join([first, second])
            if anchor == "":
                anchor = fullname
            elif anchor == "-":
                anchor = self._searchindex["objnames"][objname_idx][1] + "-" + fullname

            link = link + "#" + anchor

        return link, type_

    def resolve(self, cobj, this_url, return_type=False):
        """Resolve the link to the documentation, returns None if not found.

        Parameters
        ----------
        cobj : Dict[str, Any]
            Dict with information about the "code object" for which we are
            resolving a link.

                - cobj['name'] : function or class name (str)
                - cobj['module'] : module name (str)
                - cobj['module_short'] : shortened module name (str)
                - cobj['is_class'] : whether object is class (bool)
                - cobj['is_explicit'] : whether object is an explicit
                  backreference (referred to by sphinx markup) (bool)
        this_url: str
            URL of the current page. Needed to construct relative URLs
            (only used if relative=True in constructor).
        return_type : bool
            If True, return the type as well.

        Returns
        -------
        link : str or None
            The link (URL) to the documentation.
        type_ : str
            The type. Only returned if return_type is True.
        """
        full_name = cobj["module_short"] + "." + cobj["name"]
        if full_name not in self._link_cache:
            # we don't have it cached
            use_full_module = False
            for pattern in self.config["prefer_full_module"]:
                if re.search(pattern, cobj["module"] + "." + cobj["name"]):
                    use_full_module = True
                    break
            self._link_cache[full_name] = self._get_link_type(cobj, use_full_module)
        link, type_ = self._link_cache[full_name]

        if self.relative and link is not None:
            link = os.path.relpath(link, start=this_url)
            if self._is_windows:
                # replace '\' with '/' so it on the web
                link = link.replace("\\", "/")

            # for some reason, the relative link goes one directory too high up
            link = link[3:]

        return (link, type_) if return_type else link


def _handle_http_url_error(e, msg="fetching"):
    if isinstance(e, HTTPError):
        error_msg = f"{msg} {e.url}: {e.code} ({e.msg})"
    elif isinstance(e, URLError):
        error_msg = f"{msg}: {e.reason}"
    logger.warning(
        "The following {} has occurred {}".format(type(e).__name__, error_msg)
    )


def _sanitize_css_class(s):
    for x in "~!@$%^&*()+=,./';:\"?><[]\\{}|`#":
        s = s.replace(x, "-")
    return s


def _get_intersphinx_inventory(app):
    """
    Get the mapping between module names and intersphinx inventories.

    In some cases, intersphinx inventories may provide documentation for modules _other_
    than the one given by the intersphinx_mapping key (e.g., the Python inventory
    contains documentation for all standard library modules, or Matplotlib contains
    several `mpl_toolkits`` modules), so this checks py:module for all inventories and
    adds that additional module mapping.
    """
    if inventory := getattr(app.env, "sg_intersphinx_inventory", None):
        return inventory

    # Make a copy of the inventories, because this dict is created by intersphinx and we
    # don't want to break whatever assumptions it has made about it.
    intersphinx_inv_orig = getattr(app.env, "intersphinx_named_inventory", dict())
    intersphinx_inv = intersphinx_inv_orig.copy()
    for module_name, inventory in intersphinx_inv_orig.items():
        documented_modules = {
            qualname.split(".")[0] for qualname in inventory.get("py:module", dict())
        }
        for other_module_name in documented_modules - {module_name}:
            intersphinx_inv[other_module_name] = inventory

    app.env.sg_intersphinx_inventory = intersphinx_inv
    return intersphinx_inv


def _embed_code_links(app, gallery_conf, gallery_dir):
    """Add resolvers for the packages for which we want to show links."""
    doc_resolvers = {}

    src_gallery_dir = os.path.join(app.builder.srcdir, gallery_dir)
    for this_module, url in gallery_conf["reference_url"].items():
        try:
            if url is None:
                doc_resolvers[this_module] = SphinxDocLinkResolver(
                    app.config.sphinx_gallery_conf,
                    Path(app.builder.outdir),
                    src_gallery_dir,
                    relative=True,
                )
            else:
                doc_resolvers[this_module] = SphinxDocLinkResolver(
                    app.config.sphinx_gallery_conf, url, src_gallery_dir
                )

        except (URLError, HTTPError) as e:
            _handle_http_url_error(e)

    html_gallery_dir = os.path.abspath(os.path.join(app.builder.outdir, gallery_dir))

    # patterns for replacement
    link_pattern = '<a href="{link}" title="{title}" class="{css_class}">{text}</a>'
    orig_pattern = '<span class="n">%s</span>'
    period = '<span class="o">.</span>'

    # This could be turned into a generator if necessary, but should be okay
    flat = [
        [dirpath, filename]
        for dirpath, _, filenames in os.walk(html_gallery_dir)
        for filename in filenames
        if filename.endswith(".html")
    ]
    iterator = status_iterator(
        flat,
        f"embedding documentation hyperlinks for {gallery_dir}... ",
        color="fuchsia",
        length=len(flat),
        stringify_func=lambda x: os.path.basename(x[1]),
    )
    intersphinx_inv = _get_intersphinx_inventory(app)
    for dirpath, fname in iterator:
        full_fname = os.path.join(html_gallery_dir, dirpath, fname)
        subpath = dirpath[len(html_gallery_dir) + 1 :]
        json_fname = os.path.join(
            src_gallery_dir, subpath, fname[:-5] + ".codeobj.json"
        )
        if not os.path.exists(json_fname):
            continue

        # we have a json file with the objects to embed links for
        example_code_obj = _read_json(json_fname)

        # generate replacement strings with the links
        str_repl = {}
        for name in sorted(example_code_obj):
            cobjs = example_code_obj[name]
            # possible names from identify_names, which in turn gets
            # possibilities from NameFinder.get_mapping
            link = type_ = None
            for cobj in cobjs:
                for modname in (cobj["module_short"], cobj["module"]):
                    this_module = modname.split(".")[0]
                    cname = cobj["name"]

                    # Try doc resolvers first
                    if this_module in doc_resolvers:
                        try:
                            link, type_ = doc_resolvers[this_module].resolve(
                                cobj, full_fname, return_type=True
                            )
                        except (HTTPError, URLError) as e:
                            _handle_http_url_error(
                                e, msg=f"resolving {modname}.{cname}"
                            )

                    # next try intersphinx
                    if this_module == modname == "builtins":
                        this_module = "python"
                    if link is None and this_module in intersphinx_inv:
                        inv = intersphinx_inv[this_module]
                        if modname == "builtins":
                            want = cname
                        else:
                            want = f"{modname}.{cname}"
                        for key, value in inv.items():
                            # only python domain
                            if key.startswith("py") and want in value:
                                link = value[want][2]
                                type_ = key
                                break

                    # differentiate classes from instances
                    is_instance = (
                        type_ is not None
                        and "py:class" in type_
                        and not cobj["is_class"]
                    )

                    if link is not None:
                        # Add CSS classes
                        name_html = period.join(
                            orig_pattern % part for part in name.split(".")
                        )
                        full_function_name = f"{modname}.{cname}"
                        css_class = "sphx-glr-backref-module-" + _sanitize_css_class(
                            modname
                        )
                        if type_ is not None:
                            css_class += (
                                " sphx-glr-backref-type-" + _sanitize_css_class(type_)
                            )
                        if is_instance:
                            css_class += " sphx-glr-backref-instance"
                        str_repl[name_html] = link_pattern.format(
                            link=link,
                            title=full_function_name,
                            css_class=css_class,
                            text=name_html,
                        )
                        break  # loop over possible module names

                if link is not None:
                    break  # loop over cobjs

        # do the replacement in the html file

        # ensure greediness
        names = sorted(str_repl, key=len, reverse=True)
        regex_str = "|".join(re.escape(name) for name in names)
        regex = re.compile(regex_str)

        def substitute_link(match):
            return str_repl[match.group()]

        if len(str_repl) > 0:
            with open(full_fname, "r", encoding="utf-8") as fid:
                lines_in = fid.readlines()
            with open(full_fname, "w", **_W_KW) as fid:
                for line in lines_in:
                    line_out = regex.sub(substitute_link, line)
                    fid.write(line_out)


def embed_code_links(app, exception):
    """Embed hyperlinks to documentation into example code."""
    if exception is not None:
        return

    gallery_conf = app.config.sphinx_gallery_conf

    # XXX: Allowlist of builders for which it makes sense to embed
    # hyperlinks inside the example html. Note that the link embedding
    # require searchindex.js to exist for the links to the local doc
    # and there does not seem to be a good way of knowing which
    # builders creates a searchindex.js.
    if app.builder.name not in ["html", "readthedocs"]:
        return

    logger.info("embedding documentation hyperlinks...", color="white")

    gallery_dirs = gallery_conf["gallery_dirs"]
    if not isinstance(gallery_dirs, list):
        gallery_dirs = [gallery_dirs]

    for gallery_dir in gallery_dirs:
        _embed_code_links(app, gallery_conf, gallery_dir)



"""
.. _glyph_example:

Plotting Glyphs (Vectors or PolyData)
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Use vectors in a dataset to plot and orient glyphs/geometric objects.
"""

import numpy as np

# sphinx_gallery_thumbnail_number = 4
import pyvista as pv
from pyvista import examples

###############################################################################
# Glyphying can be done via the :meth:`pyvista.DataSetFilters.glyph` filter

mesh = examples.download_carotid().threshold(145, scalars="scalars")
mask = mesh["scalars"] < 210
mesh["scalars"][mask] = 0  # null out smaller vectors

# Make a geometric object to use as the glyph
geom = pv.Arrow()  # This could be any dataset

# Perform the glyph
glyphs = mesh.glyph(orient="vectors", scale="scalars", factor=0.003, geom=geom)

# plot using the plotting class
pl = pv.Plotter()
pl.add_mesh(glyphs, show_scalar_bar=False, lighting=False, cmap="coolwarm")
pl.camera_position = [
    (146.53, 91.28, 21.70),
    (125.00, 94.45, 19.81),
    (-0.086, 0.007, 0.996),
]  # view only part of the vector field
cpos = pl.show(return_cpos=True)


###############################################################################
# Another approach is to load the vectors directly to the mesh object and then
# access the :attr:`pyvista.DataSet.arrows` property.

sphere = pv.Sphere(radius=3.14)

# make cool swirly pattern
vectors = np.vstack(
    (
        np.sin(sphere.points[:, 0]),
        np.cos(sphere.points[:, 1]),
        np.cos(sphere.points[:, 2]),
    )
).T

# add and scale
sphere["vectors"] = vectors * 0.3
sphere.set_active_vectors("vectors")

# plot just the arrows
sphere.arrows.plot()

###############################################################################
# Plot the arrows and the sphere.

p = pv.Plotter()
p.add_mesh(sphere.arrows, lighting=False, scalar_bar_args={"title": "Vector Magnitude"})
p.add_mesh(sphere, color="grey", ambient=0.6, opacity=0.5, show_edges=False)
p.show()


###############################################################################
# Subset of Glyphs
# ++++++++++++++++
#
# Sometimes you might not want glyphs for every node in the input dataset. In
# this case, you can choose to build glyphs for a subset of the input dataset
# by using a merging tolerance. Here we specify a merging tolerance of five
# percent which equates to five percent of the bounding box's length.

# Example dataset with normals
mesh = examples.load_random_hills()

# create a subset of arrows using the glyph filter
arrows = mesh.glyph(scale="Normals", orient="Normals", tolerance=0.05)

p = pv.Plotter()
p.add_mesh(arrows, color="black")
p.add_mesh(mesh, scalars="Elevation", cmap="terrain", smooth_shading=True)
p.show()



"""
Command line arguments support
==============================

Use command line arguments to control example script.
"""

import sys

import matplotlib.pyplot as plt
import numpy as np

if len(sys.argv) > 1 and sys.argv[1] == "plot":
    fig_0, ax_0 = plt.subplots(figsize=(5, 1))

    x = np.arange(0, 10.0, 1)
    ax_0.plot(x, x**2)
    plt.show()



"""
Plot-cos
========

"""

import matplotlib.pyplot as plt
import numpy as np

x = np.linspace(0, 2 * np.pi, 100)
y = np.cos(x)

plt.plot(x, y)
plt.xlabel(r"$x$")
plt.ylabel(r"$\sin(x)$")
# To avoid matplotlib text output
plt.show()



"""
Test plot deferring
===================

This tests the ``sphinx_gallery_defer_figures`` flag.
"""

import matplotlib.pyplot as plt

# %%
# This code block should produce no plot.

plt.plot([0, 1])
plt.plot([1, 0])
# sphinx_gallery_defer_figures

# %%
# This code block should produce a plot with three lines.

plt.plot([2, 2])
plt.show()



from .nested import NestedDummyClass  # noqa: F401


class DummyClass:
    """Dummy class for testing method resolution."""

    def run(self):
        """Do nothing."""
        pass

    @property
    def prop(self):
        """Property."""
        return "Property"



# -*- coding: utf-8 -*-
"""Recommendation system generator.

Generate recommendations based on TF-IDF representation and a KNN model.
"""
# Author: Arturo Amor
# License: 3-clause BSD

import numbers
import re
from collections import defaultdict
from pathlib import Path

from .backreferences import (
    THUMBNAIL_PARENT_DIV,
    THUMBNAIL_PARENT_DIV_CLOSE,
    _thumbnail_div,
)
from .gen_rst import extract_intro_and_title
from .py_source_parser import split_code_and_text_blocks
from .utils import _replace_md5


class ExampleRecommender:
    """Compute content-based KNN-TF-IFD recommendation system.

    Parameters
    ----------
    n_examples : int, default=5
        Number of most relevant examples to display.
    min_df : float in range [0.0, 1.0] or int, default=1
        When building the vocabulary ignore terms that have a document frequency
        strictly lower than the given threshold. If float, the parameter
        represents a proportion of documents, integer represents absolute
        counts. This value is also called cut-off in the literature.
    max_df : float in range [0.0, 1.0] or int, default=1.0
        When building the vocabulary ignore terms that have a document frequency
        strictly higher than the given threshold. If float, the parameter
        represents a proportion of documents, integer represents absolute
        counts.

    Attributes
    ----------
    file_names_ : list of str
        The list of file names used for computing the similarity matrix.
        The recommended examples are chosen among this list.

    similarity_matrix_ : dense matrix
        Fitted matrix of pairwise cosine similarities.
    """

    def __init__(self, *, n_examples=5, min_df=3, max_df=0.9):
        self.n_examples = n_examples
        self.min_df = min_df
        self.max_df = max_df

    def token_freqs(self, doc):
        """Extract a dict mapping raw tokens from doc to their occurrences."""
        token_generator = (tok.lower() for tok in re.findall(r"\w+", doc))
        return self.dict_freqs(token_generator)

    @staticmethod
    def dict_freqs(doc):
        """Extract a dict mapping list of tokens to their occurrences."""
        freq = defaultdict(int)
        for tok in doc:
            freq[tok] += 1
        return freq

    def dict_vectorizer(self, data):
        """Convert a dictionary of feature occurrence frequencies into a matrix.

        Parameters
        ----------
        data : list of dict
            Each dictionary represents a document where tokens are keys and
            values are their occurrence frequencies.

        Returns
        -------
        X : ndarray of shape (n_samples, n_features)
            A matrix of occurrences where n_samples is the number of samples in
            the dataset and n_features is the total number of features across
            all samples.
        """
        import numpy as np

        tokens = []
        all_values = defaultdict(list)
        for dict_of_freqs in data:
            for token, freq in dict_of_freqs.items():
                tokens.append(token)
                all_values[token].append(freq)

        token_dict = {
            token: token_idx for token_idx, token in enumerate(sorted(all_values))
        }
        X = np.zeros((len(data), len(token_dict)))
        for dict_of_freqs_idx, dict_of_freqs in enumerate(data):
            for token, freq in dict_of_freqs.items():
                X[dict_of_freqs_idx, token_dict[token]] = freq
        return X

    def compute_tf_idf(self, X):
        """Transform a term frequency matrix into a TF-IDF matrix.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            A term frequency matrix.

        Returns
        -------
        X_tfidf : ndarray of shape (n_samples, n_features)
            A tf-idf matrix of the same shape as X.
        """
        import numpy as np

        n_samples = X.shape[0] + 1  # avoid taking log of 0

        df = np.count_nonzero(X, axis=0) + 1
        idf = np.log(n_samples / df) + 1
        X_tfidf = X * idf
        X_tfidf = (X_tfidf.T / np.linalg.norm(X_tfidf, axis=1)).T

        return X_tfidf

    def cosine_similarity(self, X, Y=None):
        """Compute the cosine similarity between two vectors X and Y.

        Parameters
        ----------
        X : ndarray of shape (n_samples_X, n_features)
            Input data.

        Y : ndarray of shape (n_samples_Y, n_features), default=None
            Input data. If `None`, the output will be the pairwise
            similarities between all samples in `X`.

        Returns
        -------
        cosine_similarity : ndarray of shape (n_samples_X, n_samples_Y)
            Cosine similarity matrix.
        """
        import numpy as np

        if Y is X or Y is None:
            Y = X

        X_normalized = X / np.linalg.norm(X)
        if X is Y:
            Y_normalized = X_normalized
        else:
            Y_normalized = Y / np.linalg.norm(Y)
        similarity = X_normalized @ Y_normalized.T

        return similarity

    def fit(self, file_names):
        """Compute the similarity matrix of a group of documents.

        Parameters
        ----------
        file_names : list or generator of file names.

        Returns
        -------
        self : object
            Fitted recommender.
        """
        import numpy as np

        n_examples = self.n_examples
        min_df = self.min_df
        max_df = self.max_df
        if not isinstance(n_examples, numbers.Integral):
            raise ValueError("n_examples must be an integer")
        elif n_examples < 1:
            raise ValueError("n_examples must be strictly positive")
        if not (
            (isinstance(min_df, numbers.Integral) and min_df >= 1)
            or (isinstance(min_df, float) and 0.0 <= min_df <= 1.0)
        ):
            raise ValueError("min_df must be float in range [0.0, 1.0] or int")
        if not (
            (isinstance(max_df, numbers.Integral) and max_df > 1)
            or (isinstance(max_df, float) and 0.0 <= max_df <= 1.0)
        ):
            raise ValueError("max_df must be float in range [0.0, 1.0] or int")

        freq_func = self.token_freqs
        counts_matrix = self.dict_vectorizer(
            [freq_func(Path(fname).read_text(encoding="utf-8")) for fname in file_names]
        )
        if isinstance(min_df, float):
            min_df = int(np.ceil(min_df * counts_matrix.shape[0]))
        if isinstance(max_df, float):
            max_df = int(np.floor(max_df * counts_matrix.shape[0]))
        doc_appearances = np.sum(counts_matrix, axis=0)
        mask = (doc_appearances >= min_df) & (doc_appearances <= max_df)
        self.similarity_matrix_ = self.cosine_similarity(
            self.compute_tf_idf(counts_matrix[:, mask])
        )
        self.file_names_ = file_names
        return self

    def predict(self, file_name):
        """Compute the `n_examples` most similar documents to the query.

        Parameters
        ----------
        file_name : str
            Name of the file corresponding to the query index `item_id`.

        Returns
        -------
        recommendations : list of str
            Name of the files most similar to the query.
        """
        item_id = self.file_names_.index(file_name)
        similar_items = list(enumerate(self.similarity_matrix_[item_id]))
        sorted_items = sorted(similar_items, key=lambda x: x[1], reverse=True)

        # Get the top k items similar to item_id. Note that `sorted_items[0]`
        # is always the query document itself, hence it is discarded from the
        # returned list of recommendations.
        top_k_items = [idx for idx, _ in sorted_items[1 : self.n_examples + 1]]
        recommendations = [self.file_names_[idx] for idx in top_k_items]
        return recommendations


def _write_recommendations(recommender, fname, gallery_conf):
    """Generate `.recommendations` reST file for a given example.

    Parameters
    ----------
    recommender : ExampleRecommender
        Instance of a fitted ExampleRecommender.

    fname : str
        Path to the example file.

    gallery_conf : dict
        Configuration dictionary for the sphinx-gallery extension.
    """
    path_fname = Path(fname)
    recommend_fname = f"{path_fname.parent / path_fname.stem}.recommendations.new"
    recommended_examples = recommender.predict(fname)

    default_rubric_header = "Related examples"
    rubric_header = gallery_conf["recommender"].get(
        "rubric_header", default_rubric_header
    )

    with open(recommend_fname, "w", encoding="utf-8") as ex_file:
        ex_file.write(f"\n\n.. rubric:: {rubric_header}\n")
        ex_file.write(THUMBNAIL_PARENT_DIV)
        for example_fname in recommended_examples:
            example_path = Path(example_fname)
            _, script_blocks = split_code_and_text_blocks(
                example_fname, return_node=False
            )
            intro, title = extract_intro_and_title(fname, script_blocks[0].content)
            ex_file.write(
                _thumbnail_div(
                    example_path.parent,
                    gallery_conf["src_dir"],
                    example_path.name,
                    intro,
                    title,
                    is_backref=True,
                )
            )
        ex_file.write(THUMBNAIL_PARENT_DIV_CLOSE)
    _replace_md5(recommend_fname, mode="t")



"""
Force images to be displayed on separate lines
==============================================
This example demonstrates how the ``sphinx_gallery_multi_image`` option can be used to
force images to be displayed on separate lines, rather than the default behaviour of
displaying them side-by-side.
"""

# sphinx_gallery_multi_image = "single"

# %%

import matplotlib.pyplot as plt
import numpy as np

fig, ax = plt.subplots(1, 1, figsize=(8, 4))
ax.pcolormesh(np.random.randn(100, 100))

fig, ax = plt.subplots(1, 1, figsize=(8, 4))
ax.pcolormesh(np.random.randn(100, 100))



"""
Plotting the exponential function
=================================

This example demonstrates how to import a local module and how images are stacked when
two plots are created in one code block (see the :doc:`plot_9_multi_image_separate`
example for information on controlling this behaviour). The variable ``N`` from the
example 'Local module' (file ``local_module.py``) is imported in the code below.
Further, note that when there is only one code block in an example, the output appears
before the code block.
"""

# Code source: Óscar Nájera
# License: BSD 3 clause

import matplotlib.pyplot as plt
import numpy as np

# You can use modules local to the example being run, here we import
# N from local_module
from local_module import N  # = 100


def main():
    """Plot exponential functions."""
    x = np.linspace(-1, 2, N)
    y = np.exp(x)

    plt.figure()
    plt.plot(x, y)
    plt.xlabel("$x$")
    plt.ylabel(r"$\exp(x)$")
    plt.title("Exponential function")

    plt.figure()
    plt.plot(x, -np.exp(-x))
    plt.xlabel("$x$")
    plt.ylabel(r"$-\exp(-x)$")
    plt.title("Negative exponential\nfunction")
    # To avoid matplotlib text output
    plt.show()


if __name__ == "__main__":
    main()



# Author: Óscar Nájera
# License: 3-clause BSD
"""Testing the rst files generator."""

import ast
import codecs
import codeop
import importlib
import io
import logging
import os
import re
import shutil
import tempfile
import zipfile
from pathlib import Path
from unittest import mock

import pytest
from sphinx.errors import ExtensionError

import sphinx_gallery.gen_rst as sg
from sphinx_gallery import downloads
from sphinx_gallery.gen_gallery import (
    _update_gallery_conf_exclude_implicit_doc,
    generate_dir_rst,
)
from sphinx_gallery.interactive_example import check_binder_conf

# TODO: The tests of this method should probably be moved to test_py_source_parser.py
from sphinx_gallery.py_source_parser import Block, split_code_and_text_blocks
from sphinx_gallery.scrapers import ImagePathIterator, figure_rst

root = Path(__file__).parents[2]

CONTENT = [
    '"""',
    "================",
    "Docstring header",
    "================",
    "",
    "This is the description of the example",
    "which goes on and on, Óscar",
    "",
    "",
    "And this is a second paragraph",
    '"""',
    "",
    "# sphinx_gallery_thumbnail_number = 1",
    "# sphinx_gallery_defer_figures",
    "# and now comes the module code",
    "# sphinx_gallery_start_ignore",
    "pass # Will be run but not rendered",
    "# sphinx_gallery_end_ignore",
    "import logging",
    "import sys",
    "from warnings import warn",
    "x, y = 1, 2",
    'print("Óscar output") # need some code output',
    "logger = logging.getLogger()",
    "logger.setLevel(logging.INFO)",
    "lh = logging.StreamHandler(sys.stdout)",
    'lh.setFormatter(logging.Formatter("log:%(message)s"))',
    "logger.addHandler(lh)",
    'logger.info("Óscar")',
    'print(r"$\\langle n_\\uparrow n_\\downarrow \\rangle$")',
    'warn("WarningsAbound", RuntimeWarning)',
]


def test_split_code_and_text_blocks():
    """Test if a known example file gets properly split."""
    file_conf, blocks = split_code_and_text_blocks(
        root / "examples" / "no_output" / "just_code.py",
    )

    assert file_conf == {}
    assert blocks[0][0] == "text"
    assert blocks[1][0] == "code"


def test_bug_cases_of_notebook_syntax():
    """Test the block splitting with supported syntax in notebook styled example.

    `plot_parse.py` uses both '#'s' and '#%%' as cell separators.
    """
    ref_blocks = ast.literal_eval(
        (Path(__file__).parent / "reference_parse.txt").read_text("utf-8")
    )
    file_conf, blocks = split_code_and_text_blocks(root / "tutorials" / "plot_parse.py")

    assert file_conf == {}
    assert blocks == ref_blocks


def test_direct_comment_after_docstring():
    # For more details see
    # https://github.com/sphinx-gallery/sphinx-gallery/pull/49
    with tempfile.NamedTemporaryFile("w", delete=False) as f:
        f.write(
            "\n".join(
                [
                    '"Docstring"',
                    "# and now comes the module code",
                    "# with a second line of comment",
                    "x, y = 1, 2",
                    "",
                ]
            )
        )
    try:
        file_conf, result = split_code_and_text_blocks(f.name)
    finally:
        os.remove(f.name)

    assert file_conf == {}
    expected_result = [
        ("text", "Docstring", 1),
        (
            "code",
            "\n".join(
                [
                    "# and now comes the module code",
                    "# with a second line of comment",
                    "x, y = 1, 2",
                    "",
                ]
            ),
            2,
        ),
    ]
    assert result == expected_result


def test_final_rst_last_word(tmpdir):
    """Tests last word in final rst block included as text."""
    filename = str(tmpdir.join("temp.py"))
    with open(filename, "w") as f:
        f.write(
            "\n".join(
                [
                    '"Docstring"',
                    "# comment only code block",
                    "#%%",
                    "# Include this whole sentence.",
                ]
            )
        )

    file_conf, result = split_code_and_text_blocks(f.name)

    assert file_conf == {}
    expected_result = [
        ("text", "Docstring", 1),
        ("code", "# comment only code block\n", 2),
        ("text", "Include this whole sentence.", 4),
    ]
    assert result == expected_result


def test_rst_block_after_docstring(gallery_conf, tmpdir):
    """Assert there is a blank line between the docstring and rst blocks."""
    filename = str(tmpdir.join("temp.py"))
    with open(filename, "w") as f:
        f.write(
            "\n".join(
                [
                    '"Docstring"',
                    "####################",
                    "# Paragraph 1",
                    "# is long.",
                    "",
                    "#%%",
                    "# Paragraph 2",
                    "",
                    "# %%",
                    "# Paragraph 3",
                    "",
                ]
            )
        )
    file_conf, blocks = split_code_and_text_blocks(filename)

    assert file_conf == {}
    assert len(blocks) == 4
    assert blocks[0][0] == "text"
    assert blocks[1][0] == "text"
    assert blocks[2][0] == "text"
    assert blocks[3][0] == "text"

    script_vars = {"execute_script": ""}
    file_conf = {}

    output_blocks, time_elapsed = sg.execute_script(
        blocks, script_vars, gallery_conf, file_conf
    )

    example_rst = sg.rst_blocks(blocks, output_blocks, file_conf, gallery_conf)
    want_rst = """\
Docstring

.. GENERATED FROM PYTHON SOURCE LINES 3-5

Paragraph 1
is long.

.. GENERATED FROM PYTHON SOURCE LINES 7-8

Paragraph 2

.. GENERATED FROM PYTHON SOURCE LINES 10-11

Paragraph 3

"""
    assert example_rst == want_rst


def test_rst_empty_code_block(gallery_conf, tmpdir):
    """Test that we can "execute" a code block containing only comments."""
    gallery_conf.update(image_scrapers=())
    filename = str(tmpdir.join("temp.py"))
    with open(filename, "w") as f:
        f.write(
            "\n".join(
                [
                    '"Docstring"',
                    "####################",
                    "# Paragraph 1",
                    "",
                    "# just a comment",
                ]
            )
        )
    file_conf, blocks = split_code_and_text_blocks(filename)

    assert file_conf == {}
    assert len(blocks) == 3
    assert blocks[0][0] == "text"
    assert blocks[1][0] == "text"
    assert blocks[2][0] == "code"

    gallery_conf["abort_on_example_error"] = True
    script_vars = dict(
        execute_script=True,
        src_file=filename,
        image_path_iterator=[],
        target_file=filename,
    )

    output_blocks, time_elapsed = sg.execute_script(
        blocks, script_vars, gallery_conf, file_conf
    )

    example_rst = sg.rst_blocks(blocks, output_blocks, file_conf, gallery_conf)
    want_rst = """\
Docstring

.. GENERATED FROM PYTHON SOURCE LINES 3-4

Paragraph 1

.. GENERATED FROM PYTHON SOURCE LINES 4-5

.. code-block:: python


    # just a comment"""
    assert example_rst.rstrip("\n") == want_rst


def test_script_vars_globals(gallery_conf, tmpdir):
    """Assert the global vars get stored."""
    gallery_conf.update(image_scrapers=())
    filename = str(tmpdir.join("temp.py"))
    with open(filename, "w") as f:
        f.write(
            """
'''
My example
----------

This is it.
'''
a = 1.
b = 'foo'
"""
        )
    file_conf, blocks = split_code_and_text_blocks(filename)
    assert len(blocks) == 2
    assert blocks[0][0] == "text"
    assert blocks[1][0] == "code"
    assert file_conf == {}
    script_vars = {
        "execute_script": True,
        "src_file": filename,
        "image_path_iterator": [],
        "target_file": filename,
    }
    output_blocks, time_elapsed = sg.execute_script(
        blocks, script_vars, gallery_conf, file_conf
    )
    assert "example_globals" in script_vars
    assert script_vars["example_globals"]["a"] == 1.0
    assert script_vars["example_globals"]["b"] == "foo"


def test_codestr2rst():
    """Test the correct translation of a code block into rst."""
    output = sg.codestr2rst('print("hello world")')
    reference = """\
.. code-block:: python

    print("hello world")"""
    assert reference == output


def test_extract_intro_and_title():
    intro, title = sg.extract_intro_and_title("<string>", "\n".join(CONTENT[1:10]))
    assert title == "Docstring header"
    assert "Docstring" not in intro
    assert intro == "This is the description of the example which goes on and on, Óscar"  # noqa
    assert "second paragraph" not in intro

    # SG incorrectly grabbing description when a label is defined (gh-232)
    intro_label, title_label = sg.extract_intro_and_title(
        "<string>", "\n".join([".. my_label", ""] + CONTENT[1:10])
    )
    assert intro_label == intro
    assert title_label == title

    intro_whitespace, title_whitespace = sg.extract_intro_and_title(
        "<string>", "\n".join(CONTENT[1:4] + [""] + CONTENT[5:10])
    )
    assert intro_whitespace == intro
    assert title_whitespace == title

    # Make example title optional (gh-222)
    intro, title = sg.extract_intro_and_title("<string>", "Title\n-----")
    assert intro == title == "Title"

    # Title beginning with a space (gh-356)
    intro, title = sg.extract_intro_and_title(
        "filename", "^^^^^\n   Title  two  \n^^^^^"
    )
    assert intro == title == "Title  two"

    # Title with punctuation (gh-517)
    intro, title = sg.extract_intro_and_title(
        "<string>",
        '    ------------\n"-`Header"-with:; `punct` mark\'s\n----------------',
    )  # noqa: E501
    assert title == '"-`Header"-with:; punct mark\'s'

    # Long intro paragraph are preserved
    intro_paragraph = "\n".join(["this is one line" for _ in range(10)])
    intro, _ = sg.extract_intro_and_title(
        "filename", "Title\n-----\n\n" + intro_paragraph
    )
    assert len(intro_paragraph) > 100
    assert len(intro) > 100
    assert intro_paragraph.replace("\n", " ") == intro

    # Errors
    with pytest.raises(ExtensionError, match="should have a header"):
        sg.extract_intro_and_title("<string>", "")  # no title
    with pytest.raises(ExtensionError, match="Could not find a title"):
        sg.extract_intro_and_title("<string>", "=====")  # no real title


@pytest.mark.parametrize(
    "mode,expected_md5",
    (
        ["b", "a546be453c8f436e744838a4801bd3a0"],
        ["t", "ea8a570e9f3afc0a7c3f2a17a48b8047"],
    ),
)
def test_md5sums(mode, expected_md5, tmp_path):
    """Test md5sum check functions work on know file content."""
    file_content = b"Local test\r\n"
    fname = tmp_path / "test"
    fname.write_bytes(file_content)
    file_md5 = sg.get_md5sum(fname, mode)
    # verify correct md5sum
    assert file_md5 == expected_md5, mode
    # False because is a new file
    assert not sg.md5sum_is_current(fname), mode
    # Write md5sum to file to check is current
    with open(str(fname) + ".md5", "w") as file_checksum:
        file_checksum.write(file_md5)
    assert sg.md5sum_is_current(fname, mode), mode


@pytest.mark.parametrize(
    "failing_code, want",
    [
        (
            CONTENT + ["#" * 79, "First_test_fail", "#" * 79, "second_fail"],
            "not defined",
        ),
        (
            CONTENT + ["#" * 79, 'input("foo")', "#" * 79, "second_fail"],
            "Cannot use input",
        ),
        (CONTENT + ["#" * 79, "bad syntax", "#" * 79, "second_fail"], "invalid syntax"),
    ],
)
def test_fail_example(gallery_conf, failing_code, want, log_collector, req_pil):
    """Test that failing examples are only executed until failing block."""
    gallery_conf.update(image_scrapers=(), reset_modules=())
    gallery_conf.update(filename_pattern="raise.py")

    with codecs.open(
        os.path.join(gallery_conf["examples_dir"], "raise.py"),
        mode="w",
        encoding="utf-8",
    ) as f:
        f.write("\n".join(failing_code))

    sg.generate_file_rst(
        "raise.py",
        gallery_conf["gallery_dir"],
        gallery_conf["examples_dir"],
        gallery_conf,
    )
    log_collector.warning.assert_called_once()
    msg = log_collector.warning.call_args[0][2]
    assert want in msg
    assert "gen_gallery" not in msg
    # can only check that gen_rst is removed on non-input ones
    if "Cannot use input" not in msg:
        assert "gen_rst" not in msg
    assert "_check_input" not in msg

    # read rst file and check if it contains traceback output

    with codecs.open(
        os.path.join(gallery_conf["gallery_dir"], "raise.rst"),
        mode="r",
        encoding="utf-8",
    ) as f:
        ex_failing_blocks = f.read().count("pytb")
        assert ex_failing_blocks != 0, "Did not run into errors in bad code"
        assert ex_failing_blocks <= 1, "Did not stop executing script after error"


def _generate_rst(gallery_conf, fname, content):
    """Return the reST text of a given example content.

    This writes a file gallery_conf['examples_dir']/fname with *content*,
    creates the corresponding rst file by running generate_file_rst() and
    returns the generated reST code.

    Parameters
    ----------
    gallery_conf
        A gallery_conf as created by the gallery_conf fixture.
    fname : str
        A filename; e.g. 'test.py'. This is relative to
        gallery_conf['examples_dir']
    content : str
        The content of fname.

    Returns
    -------
    rst : str
        The generated reST code.
    """
    with codecs.open(
        os.path.join(gallery_conf["examples_dir"], fname), mode="w", encoding="utf-8"
    ) as f:
        f.write("\n".join(content))
    with codecs.open(
        os.path.join(gallery_conf["examples_dir"], "README.txt"), "w", "utf8"
    ):
        pass

    # generate rst file
    generate_dir_rst(
        gallery_conf["examples_dir"],
        gallery_conf["gallery_dir"],
        gallery_conf,
        set(),
        is_subsection=False,
    )
    # read rst file and check if it contains code output
    rst_fname = os.path.splitext(fname)[0] + ".rst"
    with codecs.open(
        os.path.join(gallery_conf["gallery_dir"], rst_fname), mode="r", encoding="utf-8"
    ) as f:
        rst = f.read()
    return rst


ALPHA_CONTENT = '''
"""
Make a plot
===========

Plot.
"""
import matplotlib.pyplot as plt
plt.plot([0, 1], [0, 1])
'''.split("\n")


def _alpha_mpl_scraper(block, block_vars, gallery_conf):
    import matplotlib.pyplot as plt

    image_path_iterator = block_vars["image_path_iterator"]
    image_paths = list()
    for fig_num, image_path in zip(plt.get_fignums(), image_path_iterator):
        fig = plt.figure(fig_num)
        assert image_path.endswith(".png")
        # use format that does not support alpha
        image_path = Path(image_path).with_suffix(".jpg")
        fig.savefig(image_path)
        image_paths.append(image_path)
    plt.close("all")
    return figure_rst(image_paths, gallery_conf["src_dir"])


def test_custom_scraper_thumbnail_alpha(gallery_conf, req_mpl_jpg):
    """Test that thumbnails without an alpha channel work w/custom scraper."""
    gallery_conf["image_scrapers"] = [_alpha_mpl_scraper]
    rst = _generate_rst(gallery_conf, "plot_test.py", ALPHA_CONTENT)
    assert ".jpg" in rst


def test_remove_config_comments(gallery_conf, req_pil):
    """Test the gallery_conf['remove_config_comments'] setting."""
    rst = _generate_rst(gallery_conf, "test.py", CONTENT)
    assert "# sphinx_gallery_thumbnail_number = 1" in rst
    assert "# sphinx_gallery_defer_figures" in rst

    gallery_conf["remove_config_comments"] = True
    rst = _generate_rst(gallery_conf, "test.py", CONTENT)
    assert "# sphinx_gallery_thumbnail_number = 1" not in rst
    assert "# sphinx_gallery_defer_figures" not in rst


@pytest.mark.parametrize("remove_config_comments", [True, False])
def test_remove_ignore_blocks(gallery_conf, req_pil, remove_config_comments):
    """Test removal of ignore blocks."""
    gallery_conf["remove_config_comments"] = remove_config_comments
    rst = _generate_rst(gallery_conf, "test.py", CONTENT)
    assert "pass # Will be run but not rendered" in CONTENT
    assert "pass # Will be run but not rendered" not in rst
    assert "# sphinx_gallery_start_ignore" in CONTENT
    assert "# sphinx_gallery_start_ignore" not in rst


def test_dummy_image_error(gallery_conf, req_pil):
    """Test sphinx_gallery_dummy_images configuration validated."""
    content_image = CONTENT + [
        "# sphinx_gallery_dummy_images=False",
    ]
    msg = "sphinx_gallery_dummy_images setting is not an integer"
    with pytest.raises(ExtensionError, match=msg):
        _generate_rst(gallery_conf, "test.py", content_image)


def test_final_empty_block(gallery_conf, req_pil):
    """Test empty final block is removed.

    Empty final block can occur after sole config comment is removed from final block.
    """
    content_block = CONTENT + ["# %%", "", "# sphinx_gallery_line_numbers = True"]
    gallery_conf["remove_config_comments"] = True
    gallery_conf["min_reported_time"] = -1  # Force timing info to be shown
    rst = _generate_rst(gallery_conf, "test.py", content_block)
    want = "RuntimeWarning)\n\n\n.. rst-class:: sphx-glr-timing"
    assert want in rst


def test_download_link_note_only_html(gallery_conf, req_pil):
    """Test html only directive for download_link."""
    rst = _generate_rst(gallery_conf, "test.py", CONTENT)
    download_link_note = (
        ".. only:: html\n\n"
        "    .. note::\n"
        "        :class: sphx-glr-download-link-note\n\n"
    )
    assert download_link_note in rst


def test_download_link_classes(gallery_conf, req_pil):
    """Test classes for download links."""
    rst = _generate_rst(gallery_conf, "test.py", CONTENT)
    for kind in ("python", "jupyter"):
        assert "sphx-glr-download sphx-glr-download-" + kind in rst


EXCLUDE_CONTENT = '''
""":obj:`numpy.pi` :func:`numpy.sin`"""
import numpy
numpy.pi
numpy.e
'''.split("\n")


@pytest.mark.parametrize(
    "exclusion, expected",
    [
        (None, {"numpy.sin", "numpy.pi", "numpy.e"}),
        ({".*"}, {"numpy.sin", "numpy.pi"}),
        ({"pi"}, {"numpy.sin", "numpy.pi", "numpy.e"}),
        ({r"numpy\.e", "sin"}, {"numpy.sin", "numpy.pi"}),
    ],
    ids=[
        "exclude nothing (default)",
        "exclude anything (explicit backreferences only)",
        "explicit backref not shadowed by implicit one",
        "exclude implicit backref",
    ],
)
def test_exclude_implicit(gallery_conf, exclusion, expected, monkeypatch, req_pil):
    mock_write_backreferences = mock.create_autospec(sg._write_backreferences)
    monkeypatch.setattr(sg, "_write_backreferences", mock_write_backreferences)
    gallery_conf["doc_module"] = ("numpy",)
    if exclusion:
        gallery_conf["exclude_implicit_doc"] = exclusion
        _update_gallery_conf_exclude_implicit_doc(gallery_conf)
    _generate_rst(gallery_conf, "test_exclude_implicit.py", EXCLUDE_CONTENT)
    assert mock_write_backreferences.call_args.args[0] == expected


@pytest.mark.parametrize(
    "gallery_header",
    [
        "GALLERY_HEADER.rst",
        "README.txt",
        "README.rst",
    ],
)
def test_gen_dir_rst(gallery_conf, gallery_header):
    """Test gen_dir_rst."""
    with open(os.path.join(gallery_conf["src_dir"], gallery_header), "wb") as fid:
        fid.write("Testing\n=======\n\nÓscar here.".encode())
    out = generate_dir_rst(
        gallery_conf["src_dir"], gallery_conf["gallery_dir"], gallery_conf, []
    )
    assert "Óscar here" in out[1]


@pytest.mark.parametrize(
    "gallery_header",
    [
        "GALLERY_HEADER.bad",
        "README.bad",
    ],
)
def test_gen_dir_rst_invalid_header(gallery_conf, gallery_header):
    """Check `_get_gallery_header` raises error for invalid header extension."""
    with open(os.path.join(gallery_conf["src_dir"], gallery_header), "wb") as fid:
        fid.write("Testing\n=======\n\nÓscar here.".encode())
    with pytest.raises(ExtensionError, match="does not have a GALLERY_HEADER"):
        generate_dir_rst(
            gallery_conf["src_dir"], gallery_conf["gallery_dir"], gallery_conf, []
        )


def test_pattern_matching(gallery_conf, log_collector, req_pil):
    """Test if only examples matching pattern are executed."""
    gallery_conf.update(image_scrapers=(), reset_modules=())
    gallery_conf.update(filename_pattern=re.escape(os.sep) + "plot_0")

    code_output = (
        "\n .. code-block:: none\n"
        "\n"
        "    Óscar output\n"
        "    log:Óscar\n"
        "    $\\langle n_\\uparrow n_\\downarrow \\rangle$"
    )
    warn_output = "RuntimeWarning: WarningsAbound"
    # create three files in tempdir (only one matches the pattern)
    fnames = ["plot_0.py", "plot_1.py", "plot_2.py"]
    for fname in fnames:
        rst = _generate_rst(gallery_conf, fname, CONTENT)
        rst_fname = os.path.splitext(fname)[0] + ".rst"
        if re.search(
            gallery_conf["filename_pattern"],
            os.path.join(gallery_conf["gallery_dir"], rst_fname),
        ):
            assert code_output in rst
            assert warn_output in rst
        else:
            assert code_output not in rst
            assert warn_output not in rst


@pytest.mark.parametrize(
    "test_str",
    [
        "# sphinx_gallery_thumbnail_number= 2",
        "# sphinx_gallery_thumbnail_number=2",
        "#sphinx_gallery_thumbnail_number = 2",
        "    # sphinx_gallery_thumbnail_number=2",
    ],
)
def test_thumbnail_number(test_str):
    """Test correct plot used as thumbnail image."""
    with tempfile.NamedTemporaryFile("w", delete=False) as f:
        f.write("\n".join(['"Docstring"', test_str]))
    try:
        file_conf, blocks = split_code_and_text_blocks(f.name)
    finally:
        os.remove(f.name)
    assert file_conf == {"thumbnail_number": 2}


@pytest.mark.parametrize(
    "test_str",
    [
        "# sphinx_gallery_thumbnail_path= '_static/demo.png'",
        "# sphinx_gallery_thumbnail_path='_static/demo.png'",
        "#sphinx_gallery_thumbnail_path = '_static/demo.png'",
        "    # sphinx_gallery_thumbnail_path='_static/demo.png'",
    ],
)
def test_thumbnail_path(test_str):
    """Test correct plot used for thumbnail image."""
    with tempfile.NamedTemporaryFile("w", delete=False) as f:
        f.write("\n".join(['"Docstring"', test_str]))
    try:
        file_conf, blocks = split_code_and_text_blocks(f.name)
    finally:
        os.remove(f.name)
    assert file_conf == {"thumbnail_path": "_static/demo.png"}


def test_zip_python(gallery_conf):
    """Test generated zipfiles are not corrupt and have expected name and contents."""
    gallery_conf.update(examples_dir=os.path.join(gallery_conf["src_dir"], "examples"))
    shutil.copytree(
        os.path.join(os.path.dirname(__file__), "tinybuild", "examples"),
        gallery_conf["examples_dir"],
    )
    examples = downloads.list_downloadable_sources(gallery_conf["examples_dir"])
    zipfilepath = downloads.python_zip(examples, gallery_conf["gallery_dir"])
    assert zipfilepath.endswith("_python.zip")
    zipf = zipfile.ZipFile(zipfilepath)
    check = zipf.testzip()
    assert not check, f"Bad file in zipfile: {check}"
    filenames = {item.filename for item in zipf.filelist}
    assert "examples/plot_command_line_args.py" in filenames
    assert "examples/julia_sample.jl" not in filenames


def test_zip_mixed_source(gallery_conf):
    """Test generated zipfiles are not corrupt and have expected name and contents."""
    gallery_conf.update(examples_dir=os.path.join(gallery_conf["src_dir"], "examples"))
    shutil.copytree(
        os.path.join(os.path.dirname(__file__), "tinybuild", "examples"),
        gallery_conf["examples_dir"],
    )
    examples = downloads.list_downloadable_sources(
        gallery_conf["examples_dir"], (".py", ".jl")
    )
    zipfilepath = downloads.python_zip(examples, gallery_conf["gallery_dir"], None)
    zipf = zipfile.ZipFile(zipfilepath)
    check = zipf.testzip()
    assert not check, f"Bad file in zipfile: {check}"
    filenames = {item.filename for item in zipf.filelist}
    assert "examples/plot_command_line_args.py" in filenames
    assert "examples/julia_sample.jl" in filenames


def test_rst_example(gallery_conf):
    """Test generated rst file includes the correct paths for binder."""
    binder_conf = check_binder_conf(
        {
            "org": "sphinx-gallery",
            "repo": "sphinx-gallery.github.io",
            "binderhub_url": "https://mybinder.org",
            "branch": "master",
            "dependencies": "./binder/requirements.txt",
            "use_jupyter_lab": True,
        }
    )
    gallery_conf.update(binder=binder_conf)
    gallery_conf["min_reported_time"] = -1

    example_file = os.path.join(gallery_conf["gallery_dir"], "plot.py")
    sg.save_rst_example("example_rst", example_file, 0, 0, gallery_conf)

    test_file = re.sub(r"\.py$", ".rst", example_file)
    with codecs.open(test_file) as f:
        rst = f.read()

    assert "lab/tree/notebooks/plot.ipynb" in rst

    # CSS classes
    assert "rst-class:: sphx-glr-signature" in rst
    assert "rst-class:: sphx-glr-timing" in rst


@pytest.fixture(scope="function")
def script_vars(tmpdir):
    fake_main = importlib.util.module_from_spec(
        importlib.util.spec_from_loader("__main__", None)
    )
    fake_main.__dict__.update({"__doc__": ""})
    script_vars = {
        "execute_script": True,
        "image_path_iterator": ImagePathIterator(str(tmpdir.join("temp.png"))),
        "src_file": __file__,
        "memory_delta": [],
        "fake_main": fake_main,
    }
    return script_vars


def test_output_indentation(gallery_conf, script_vars):
    """Test whether indentation of code output is retained."""
    gallery_conf.update(image_scrapers=())
    compiler = codeop.Compile()

    test_string = r"\n".join(["  A B", "A 1 2", "B 3 4"])
    code = "print('" + test_string + "')"
    code_block = Block("code", code, 1)
    file_conf = {}
    output = sg.execute_code_block(
        compiler, code_block, None, script_vars, gallery_conf, file_conf
    )
    output_test_string = "\n".join(
        [line[4:] for line in output.strip().split("\n")[-3:]]
    )
    assert output_test_string == test_string.replace(r"\n", "\n")


def test_output_no_ansi(gallery_conf, script_vars):
    """Test ANSI characters are removed.

    See: https://en.wikipedia.org/wiki/ANSI_escape_code
    """
    gallery_conf.update(image_scrapers=())
    compiler = codeop.Compile()

    code = 'print("\033[94m0.25")'
    code_block = Block("code", code, 1)
    file_conf = {}
    output = sg.execute_code_block(
        compiler, code_block, None, script_vars, gallery_conf, file_conf
    )
    output_test_string = "\n".join(
        [line[4:] for line in output.strip().split("\n")[-3:]]
    )

    assert output_test_string.split("\n")[-1] == "0.25"


def test_absl_logging(gallery_conf, script_vars):
    """Test using absl logging does not throw error.

    This is important, as many popular libraries like Tensorflow use absl for logging.
    """
    pytest.importorskip("absl")

    gallery_conf.update(image_scrapers=())
    compiler = codeop.Compile()
    code = 'from absl import logging\nlogging.info("Should not crash!")'
    sg.execute_code_block(
        compiler, Block("code", code, 1), None, script_vars, gallery_conf, {}
    )

    # Previously, absl crashed during shutdown, after all tests had run, so
    # tests would pass even if this issue occurred. To address this issue, we
    # call shutdown inside the test.
    logging.shutdown()


def test_empty_output_box(gallery_conf, script_vars):
    """Tests that `print(__doc__)` doesn't produce an empty output box."""
    gallery_conf.update(image_scrapers=())
    compiler = codeop.Compile()

    code_block = Block("code", "print(__doc__)", 1)
    file_conf = {}

    output = sg.execute_code_block(
        compiler, code_block, None, script_vars, gallery_conf, file_conf
    )
    assert output.isspace()


code_repr_only = """
class repr_only_class():

    def __init__(self):
        pass

    def __repr__(self):
        return "This is the __repr__"

class_inst = repr_only_class()
class_inst
"""

code_repr_and_html = """
class repr_and_html_class():
    def __init__(self):
        pass

    def __repr__(self):
        return "This is the __repr__"

    def _repr_html_(self):
        return "<div> This is the _repr_html_ div </div>"

class_inst = repr_and_html_class()
class_inst
"""

code_print_and_repr_and_html = """
print("print statement")

class repr_and_html_class():
    def __init__(self):
        pass

    def __repr__(self):
        return "This is the __repr__"

    def _repr_html_(self):
        return "<div> This is the _repr_html_ div </div>"

class_inst = repr_and_html_class()
class_inst
"""

code_plt = """
import matplotlib.pyplot as plt
fig = plt.figure()
plt.close('all')
fig
"""

html_out = """.. raw:: html

    <div class="output_subarea output_html rendered_html output_result">
    <div> This is the _repr_html_ div </div>
    </div>
    <br />
    <br />"""

text_above_html = """.. code-block:: none

    print statement


"""


def _clean_output(output):
    is_text = ".. rst-class:: sphx-glr-script-out" in output

    is_html = ".. raw:: html" in output

    if output.isspace():
        return ""
    elif is_text and is_html:
        output_test_string = "\n".join(output.strip().split("\n")[2:])
        return output_test_string.strip()
    elif is_text:
        output_test_string = "\n".join(
            [line[4:] for line in output.strip().split("\n")[4:]]
        )
        return output_test_string.strip()
    elif is_html:
        output_test_string = "\n".join(output.strip().split("\n"))
        return output_test_string


@pytest.mark.parametrize(
    "capture_repr, code, expected_out",
    [
        pytest.param(tuple(), "a=2\nb=3", "", id="assign,()"),
        pytest.param(tuple(), "a=2\na", "", id="var,()"),
        pytest.param(tuple(), "a=2\nprint(a)", "2", id="print(var),()"),
        pytest.param(tuple(), 'print("hello")\na=2\na', "hello", id="print+var,()"),
        pytest.param(("__repr__",), "a=2\nb=3", "", id="assign,(repr)"),
        pytest.param(("__repr__",), "a=2\na", "2", id="var,(repr)"),
        pytest.param(("__repr__",), "a=2\nprint(a)", "2", id="print(var),(repr)"),
        pytest.param(
            ("__repr__",), 'print("hello")\na=2\na', "hello\n\n2", id="print+var,(repr)"
        ),
        pytest.param(
            ("__repr__",),
            code_repr_and_html,
            "This is the __repr__",
            id="repr_and_html,(repr)",
        ),
        pytest.param(
            ("__repr__",),
            code_print_and_repr_and_html,
            "print statement\n\nThis is the __repr__",
            id="print and repr_and_html,(repr)",
        ),
        pytest.param(("_repr_html_",), code_repr_only, "", id="repr_only,(html)"),
        pytest.param(
            ("_repr_html_",), code_repr_and_html, html_out, id="repr_and_html,(html)"
        ),
        pytest.param(
            ("_repr_html_",),
            code_print_and_repr_and_html,
            "".join([text_above_html, html_out]),
            id="print and repr_and_html,(html)",
        ),
        pytest.param(
            ("_repr_html_", "__repr__"),
            code_repr_and_html,
            html_out,
            id="repr_and_html,(html,repr)",
        ),
        pytest.param(
            ("__repr__", "_repr_html_"),
            code_repr_and_html,
            "This is the __repr__",
            id="repr_and_html,(repr,html)",
        ),
        pytest.param(
            ("_repr_html_", "__repr__"),
            code_repr_only,
            "This is the __repr__",
            id="repr_only,(html,repr)",
        ),
        pytest.param(("_repr_html_",), code_plt, "", id="html_none"),
    ],
)
def test_capture_repr(
    gallery_conf, capture_repr, code, expected_out, req_mpl, req_pil, script_vars
):
    """Tests output capturing with various capture_repr settings."""
    compiler = codeop.Compile()
    code_block = Block("code", code, 1)
    gallery_conf["capture_repr"] = capture_repr
    file_conf = {}
    output = sg.execute_code_block(
        compiler, code_block, None, script_vars, gallery_conf, file_conf
    )
    assert _clean_output(output) == expected_out


@pytest.mark.parametrize(
    "caprepr_gallery, caprepr_file, caprepr_block, expected_out",
    [
        pytest.param(tuple(), "()", '("__repr__",)', "2", id="() --> () --> repr"),
        pytest.param(tuple(), ("__repr__",), None, "2", id="() --> repr --> None"),
        pytest.param(("__repr__",), "()", None, "", id="repr --> () --> None"),
        pytest.param(tuple(), ("__repr__",), "()", "", id="() --> repr --> ()"),
        pytest.param(("__repr__",), "()", "()", "", id="repr --> () --> ()"),
    ],
)
def test_capture_repr_per_file_and_per_block(
    gallery_conf,
    caprepr_gallery,
    caprepr_file,
    caprepr_block,
    expected_out,
    req_mpl,
    req_pil,
    script_vars,
):
    """Tests that per file/block capture_repr overrides gallery_conf."""
    compiler = codeop.Compile()
    caprepr_block = (
        f"# sphinx_gallery_capture_repr_block={caprepr_block}\n"
        if caprepr_block
        else ""
    )
    code_block = Block("code", f"{caprepr_block}a=2\n2", 1)
    gallery_conf["capture_repr"] = caprepr_gallery
    file_conf = {"capture_repr": caprepr_file}
    output = sg.execute_code_block(
        compiler, code_block, None, script_vars, gallery_conf, file_conf
    )
    assert _clean_output(output) == expected_out


def test_ignore_repr_types(gallery_conf, req_mpl, req_pil, script_vars):
    """Tests that `ignore_repr_types` correctly ignores the output of a specific type."""
    compiler = codeop.Compile()
    code_block = Block("code", "a=2\na", 1)
    gallery_conf["ignore_repr_types"] = r"int"
    file_conf = {}
    output = sg.execute_code_block(
        compiler, code_block, None, script_vars, gallery_conf, file_conf
    )
    assert _clean_output(output) == ""


@pytest.mark.parametrize(
    ("order", "call_count"), [("before", 1), ("after", 1), ("both", 2)]
)
def test_reset_module_order_2_param(gallery_conf, order, call_count, req_pil):
    """Test that reset module with 2 parameters."""

    def cleanup_2_param(gallery_conf, fname):
        pass

    mock_reset_module = mock.create_autospec(cleanup_2_param)
    gallery_conf["reset_modules"] = (mock_reset_module,)
    gallery_conf["reset_modules_order"] = order
    _generate_rst(gallery_conf, "plot_test.py", CONTENT)
    assert mock_reset_module.call_count == call_count


@pytest.mark.parametrize(
    ("order", "call_count", "expected_call_order"),
    [
        ("before", 1, ("before",)),
        ("after", 1, ("after",)),
        ("both", 2, ("before", "after")),
    ],
)
def test_reset_module_order_3_param(
    gallery_conf, order, call_count, expected_call_order, req_pil
):
    """Test reset module with 3 parameters."""

    def cleanup_3_param(gallery_conf, fname, when):
        pass

    mock_reset_module = mock.create_autospec(cleanup_3_param)
    gallery_conf["reset_modules"] = (mock_reset_module,)
    gallery_conf["reset_modules_order"] = order
    _generate_rst(gallery_conf, "plot_test.py", CONTENT)
    assert mock_reset_module.call_count == call_count

    expected_calls = [
        mock.call(mock.ANY, mock.ANY, order) for order in expected_call_order
    ]
    mock_reset_module.assert_has_calls(expected_calls)


def test_reset_module_order_3_param_invalid_when(gallery_conf):
    """Test reset module with unknown 3rd parameter."""

    def cleanup_3_param(gallery_conf, fname, invalid):
        pass

    mock_reset_module = mock.create_autospec(cleanup_3_param)
    gallery_conf["reset_modules"] = (mock_reset_module,)
    gallery_conf["reset_modules_order"] = "before"
    with pytest.raises(
        ValueError,
        match=("3rd parameter in cleanup_3_param function signature must be 'when'"),
    ):
        _generate_rst(gallery_conf, "plot_test.py", CONTENT)
    assert mock_reset_module.call_count == 0


@pytest.fixture
def log_collector_wrap(log_collector):
    """Wrap our log_collector."""
    src_filename = "source file name"
    tee = sg._LoggingTee(src_filename)
    output_file = tee.output
    yield log_collector, src_filename, tee, output_file


def test_full_line(log_collector_wrap):
    # A full line is output immediately.
    log_collector, src_filename, tee, output_file = log_collector_wrap
    tee.write("Output\n")
    tee.flush()
    assert output_file.getvalue() == "Output\n"
    assert log_collector.verbose.call_count == 2
    assert src_filename in log_collector.verbose.call_args_list[0][0]
    assert "Output" in log_collector.verbose.call_args_list[1][0]


def test_incomplete_line_with_flush(log_collector_wrap):
    # An incomplete line ...
    log_collector, src_filename, tee, output_file = log_collector_wrap
    tee.write("Output")
    assert output_file.getvalue() == "Output"
    log_collector.verbose.assert_called_once()
    assert src_filename in log_collector.verbose.call_args[0]

    # ... should appear when flushed.
    tee.flush()
    assert log_collector.verbose.call_count == 2
    assert "Output" in log_collector.verbose.call_args_list[1][0]


def test_incomplete_line_with_more_output(log_collector_wrap):
    # An incomplete line ...
    log_collector, src_filename, tee, output_file = log_collector_wrap
    tee.write("Output")
    assert output_file.getvalue() == "Output"
    log_collector.verbose.assert_called_once()
    assert src_filename in log_collector.verbose.call_args[0]

    # ... should appear when more data is written.
    tee.write("\nMore output\n")
    assert output_file.getvalue() == "Output\nMore output\n"
    assert log_collector.verbose.call_count == 3
    assert "Output" in log_collector.verbose.call_args_list[1][0]
    assert "More output" in log_collector.verbose.call_args_list[2][0]


def test_multi_line(log_collector_wrap):
    log_collector, src_filename, tee, output_file = log_collector_wrap
    tee.write("first line\rsecond line\nthird line")
    assert output_file.getvalue() == "first line\rsecond line\nthird line"
    assert log_collector.verbose.call_count == 3
    assert src_filename in log_collector.verbose.call_args_list[0][0]
    assert "first line" in log_collector.verbose.call_args_list[1][0]
    assert "second line" in log_collector.verbose.call_args_list[2][0]
    assert tee.logger_buffer == "third line"


def test_textio_compat(log_collector_wrap):
    _, _, tee, _ = log_collector_wrap
    assert not tee.readable()
    assert not tee.seekable()
    assert tee.writable()


def test_fileno(log_collector_wrap):
    _, _, tee, _ = log_collector_wrap
    with pytest.raises(io.UnsupportedOperation):
        tee.fileno()


def test_isatty(monkeypatch, log_collector_wrap):
    _, _, tee, _ = log_collector_wrap
    assert not tee.isatty()
    monkeypatch.setattr(tee.output, "isatty", lambda: True)
    assert tee.isatty()


def test_tell(monkeypatch, log_collector_wrap):
    _, _, tee, _ = log_collector_wrap
    assert tee.tell() == tee.output.tell()
    monkeypatch.setattr(tee.output, "tell", lambda: 42)
    assert tee.tell() == tee.output.tell()


def test_errors(log_collector_wrap):
    _, _, tee, _ = log_collector_wrap
    assert tee.errors == tee.output.errors


def test_newlines(log_collector_wrap):
    _, _, tee, _ = log_collector_wrap
    assert tee.newlines == tee.output.newlines


# TODO: test that examples are executed after a no-plot and produce
#       the correct image in the thumbnail



"""Utilities.

Miscellaneous utilities.
"""

# Author: Eric Larson
# License: 3-clause BSD

import hashlib
import json
import os
import re
import subprocess
import zipfile
from collections import defaultdict
from functools import partial
from pathlib import Path
from shutil import copyfile, move

import sphinx.util
from sphinx.errors import ExtensionError

try:
    from sphinx.util.display import status_iterator  # noqa: F401
except Exception:  # Sphinx < 6
    from sphinx.util import status_iterator  # noqa: F401


logger = sphinx.util.logging.getLogger("sphinx-gallery")


# Text writing kwargs for builtins.open
_W_KW = dict(encoding="utf-8", newline="\n")


def _get_image():
    try:
        from PIL import Image
    except ImportError as exc:  # capture the error for the modern way
        try:
            import Image
        except ImportError:
            raise ExtensionError(
                "Could not import pillow, which is required "
                f"to rescale images (e.g., for thumbnails): {exc}"
            )
    return Image


def scale_image(in_fname, out_fname, max_width, max_height):
    """Scales image centered in image box using `max_width` and `max_height`.

    The same aspect ratio is retained. If `in_fname` == `out_fname` the image can only
    be scaled down.
    """
    # local import to avoid testing dependency on PIL:
    Image = _get_image()
    img = Image.open(in_fname)
    # XXX someday we should just try img.thumbnail((max_width, max_height)) ...
    width_in, height_in = img.size
    scale_w = max_width / float(width_in)
    scale_h = max_height / float(height_in)

    if height_in * scale_w <= max_height:
        scale = scale_w
    else:
        scale = scale_h

    if scale >= 1.0 and in_fname == out_fname:
        return

    width_sc = int(round(scale * width_in))
    height_sc = int(round(scale * height_in))

    # resize the image using resize; if using .thumbnail and the image is
    # already smaller than max_width, max_height, then this won't scale up
    # at all (maybe could be an option someday...)
    try:  # Pillow 9+
        bicubic = Image.Resampling.BICUBIC
    except Exception:
        bicubic = Image.BICUBIC
    img = img.resize((width_sc, height_sc), bicubic)
    # img.thumbnail((width_sc, height_sc), Image.BICUBIC)
    # width_sc, height_sc = img.size  # necessary if using thumbnail

    # insert centered
    thumb = Image.new("RGBA", (max_width, max_height), (255, 255, 255, 0))
    pos_insert = ((max_width - width_sc) // 2, (max_height - height_sc) // 2)
    thumb.paste(img, pos_insert)

    try:
        thumb.save(out_fname)
    except OSError:
        # try again, without the alpha channel (e.g., for JPEG)
        thumb.convert("RGB").save(out_fname)


def optipng(fname, args=()):
    """Optimize a PNG in place.

    Parameters
    ----------
    fname : str
        The filename. If it ends with '.png', ``optipng -o7 fname`` will
        be run. If it fails because the ``optipng`` executable is not found
        or optipng fails, the function returns.
    args : tuple
        Extra command-line arguments, such as ``['-o7']``.
    """
    fname = str(fname)
    if fname.endswith(".png"):
        # -o7 because this is what CPython used
        # https://github.com/python/cpython/pull/8032
        try:
            subprocess.check_call(
                ["optipng"] + list(args) + [fname],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
        except (subprocess.CalledProcessError, OSError):  # FileNotFoundError
            pass


def _has_optipng():
    try:
        subprocess.check_call(
            ["optipng", "--version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
    except OSError:  # FileNotFoundError
        return False
    else:
        return True


def get_md5sum(src_file, mode="b"):
    """Returns md5sum of file.

    Parameters
    ----------
    src_file : str
        Filename to get md5sum for.
    mode : 't' or 'b'
        File mode to open file with. When in text mode, universal line endings
        are used to ensure consistency in hashes between platforms.
    """
    if mode == "t":
        kwargs = {"errors": "surrogateescape", "encoding": "utf-8"}
    else:
        kwargs = {}
    # Universal newline mode is intentional here
    with open(src_file, f"r{mode}", **kwargs) as src_data:
        src_content = src_data.read()
        if mode == "t":
            src_content = src_content.encode(**kwargs)
        return hashlib.md5(src_content).hexdigest()


def _replace_md5(fname_new, fname_old=None, *, method="move", mode="b", check="md5"):
    fname_new = str(fname_new)  # convert possible Path
    assert method in ("move", "copy")
    if fname_old is None:
        assert fname_new.endswith(".new")
        fname_old = os.path.splitext(fname_new)[0]
    replace = True
    if os.path.isfile(fname_old):
        if check == "md5":  # default
            func = partial(get_md5sum, mode=mode)
        else:
            assert check == "json"

            def func(x):
                return json.loads(Path(x).read_text("utf-8"))

        try:
            equiv = func(fname_old) == func(fname_new)
        except Exception:  # e.g., old JSON file is a problem
            equiv = False
        if equiv:
            replace = False
            if method == "move":
                os.remove(fname_new)
        else:
            logger.debug(f"Replacing stale {fname_old} with {fname_new}")
    if replace:
        if method == "move":
            move(fname_new, fname_old)
        else:
            copyfile(fname_new, fname_old)
    assert os.path.isfile(fname_old)


def check_duplicate_filenames(files):
    """Check for duplicate filenames across gallery directories."""
    # Check whether we'll have duplicates
    used_names = set()
    dup_names = list()

    for this_file in files:
        this_fname = os.path.basename(this_file)
        if this_fname in used_names:
            dup_names.append(this_file)
        else:
            used_names.add(this_fname)

    if len(dup_names) > 0:
        logger.warning(
            "Duplicate example file name(s) found. Having duplicate file "
            "names will break some links. "
            "List of files: %s",
            sorted(dup_names),
        )


def check_spaces_in_filenames(files):
    """Check for spaces in filenames across example directories."""
    regex = re.compile(r"[\s]")
    files_with_space = list(filter(regex.search, files))
    if files_with_space:
        logger.warning(
            "Example file name(s) with spaces found. Having spaces in "
            "file names will break some links. "
            "List of files: %s",
            sorted(files_with_space),
        )


def _collect_gallery_files(examples_dirs, gallery_conf, check_filenames=False):
    """Collect files with `example_extensions`, accounting for `ignore_pattern`.

    If `check_filenames` we check one level of sub-folders as well as root
    `example_dirs` for gallery example files. We then check for duplicate and
    spaces in full file paths.
    """
    exts = gallery_conf["example_extensions"]
    max_depth = 1 if check_filenames else 0
    files = []
    for example_dir in examples_dirs:
        example_depth = os.path.abspath(example_dir).count(os.sep)
        for root, _, filenames in os.walk(example_dir):
            root = os.path.normpath(root)
            if (root.count(os.sep) - example_depth) > max_depth:
                break
            for filename in filenames:
                if (s := Path(filename).suffix) and s in exts:
                    if re.search(gallery_conf["ignore_pattern"], filename) is None:
                        file = filename
                        if check_filenames:
                            file = os.path.join(root, filename)
                        files.append(file)
    if check_filenames:
        check_duplicate_filenames(files)
        check_spaces_in_filenames(files)
    return files


def zip_files(file_list, zipname, relative_to, extension=None):
    """
    Creates a zip file with the given files.

    A zip file named `zipname` will be created containing the files listed in
    `file_list`. The zip file contents will be stored with their paths stripped to be
    relative to `relative_to`.
    """
    zipname_new = str(zipname) + ".new"
    with zipfile.ZipFile(zipname_new, mode="w") as zipf:
        for fname in file_list:
            if extension is not None:
                fname = os.path.splitext(fname)[0] + extension
            zipf.write(fname, os.path.relpath(fname, relative_to))
    _replace_md5(zipname_new)
    return zipname


def _has_pypandoc():
    """Check if pypandoc package available."""
    try:
        import pypandoc  # noqa

        # Import error raised only when function called
        version = pypandoc.get_pandoc_version()
    except (ImportError, OSError):
        return None, None
    else:
        return True, version


def _has_graphviz():
    try:
        import graphviz  # noqa F401
    except ImportError as exc:
        logger.info(
            "`graphviz` required for graphical visualization "
            f"but could not be imported, got: {exc}"
        )
        return False
    return True


def _escape_ansi(s):
    """Remove ANSI terminal formatting characters from a string."""
    return re.sub(r"(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]", "", s)


def _format_toctree(items, includehidden=False):
    """Format a toc tree."""
    st = """
.. toctree::
   :hidden:"""
    if includehidden:
        st += """
   :includehidden:
"""
    st += """

   {}\n""".format("\n   ".join(items))

    st += "\n"

    return st


_CUSTOM_EXAMPLE_ORDER = [
    "plot_1.py",
    "plot_3.py",
    "plot_2.py",
    "plot_5.py",
    "plot_6.py",
    "plot_4.py",
    "plot_8.py",
    "plot_7.py",
    "plot_9.py",
]


def _custom_example_sorter(filename):
    """Importable custom sorter func, used in our test suite."""
    return _CUSTOM_EXAMPLE_ORDER.index(filename)


def _custom_subsection_sorter(foldername):
    """Importable custom sorter func for subsection folders, used in our test suite."""
    return foldername[::-1]


def custom_minigallery_sort_order_sorter(file):
    """Importable custom sorter for minigallery_sort_order, used in our test suite."""
    ORDER = [
        "plot_3.py",
        "plot_2.py",
        "plot_1.py",
    ]
    return ORDER.index(Path(file).name)


# Should be matched with `_read_json`
def _write_json(target_file, to_save, name=""):
    """Write dictionary to JSON file."""
    codeobj_fname = Path(target_file).with_name(target_file.stem + f"{name}.json.new")
    with open(codeobj_fname, "w", **_W_KW) as fid:
        json.dump(
            to_save,
            fid,
            sort_keys=True,
            ensure_ascii=False,
            indent=1,
            check_circular=False,
        )
    _replace_md5(codeobj_fname, check="json")


def _read_json(json_fname):
    """Read JSON dictionary from file."""
    with open(json_fname, "r", encoding="utf-8") as fid:
        json_dict = json.load(fid)
    return json_dict


def _combine_backreferences(dict_a, dict_b):
    """Combine backreferences dictionaries, joining lists when keys are the same."""
    # `dict_b` is None when `backreferences_dir` config not set
    if isinstance(dict_b, dict):
        for key, value in dict_b.items():
            dict_a.setdefault(key, []).extend(value)
    return dict_a



"""
Identifying function names in a script
======================================

This demonstrates how Sphinx-Gallery identifies when

1. a function/method/attribute/object is used or class instantiated in a
   code block
2. a function/method/attribute/object/class is referred to using sphinx markup
   in a text block.

Sphinx-Gallery examines both the executed code itself, as well as the
text blocks (such as this one, or the one below) for these references and
identifies the module they belong to. This means that by writing
:obj:`numpy.sin` and :obj:`numpy.exp` here, they will be identified even though
they are not used in the code. This is useful in particular when functions
return classes (meaning it is not explicitly instantiated) -- if you add them
to the documented blocks of examples that use them, they will be added to
backreferences.

This functionality is used to add documentation hyperlinks to your code
(:ref:`link_to_documentation`) and for mini-gallery creation
(:ref:`references_to_examples`).
"""

# Code source: Óscar Nájera
# License: BSD 3 clause

import os.path as op  # noqa, analysis:ignore

import matplotlib.pyplot as plt

import sphinx_gallery
from sphinx_gallery.backreferences import _make_ref_regex, identify_names
from sphinx_gallery.py_source_parser import split_code_and_text_blocks

filename = "plot_6_function_identifier.py"
if not op.exists(filename):
    filename = (
        __file__
        if "__file__" in locals()
        else op.join(op.dirname(sphinx_gallery.__path__[0]), "examples", filename)
    )

_, script_blocks = split_code_and_text_blocks(filename)

names = identify_names(script_blocks, _make_ref_regex())

# %%
# In the code block above, we use the internal function ``identify_names`` to
# obtain all identified names from this file and their full resolved import
# path. We then plot this below, where the identified names functions are
# on the left and the full resolved import path is on the right.

fontsize = 12.5
figheight = 3 * len(names) * fontsize / 72
fig, ax = plt.subplots(figsize=(7.5, figheight))
ax.set_visible(False)

for i, (name, obj) in enumerate(names.items(), 1):
    fig.text(
        0.48,
        1 - i / (len(names) + 1),
        name,
        ha="right",
        va="center",
        size=fontsize,
        transform=fig.transFigure,
        bbox=dict(boxstyle="square", fc="w", ec="k"),
    )
    fig.text(
        0.52,
        1 - i / (len(names) + 1),
        obj[0]["module"],
        ha="left",
        va="center",
        size=fontsize,
        transform=fig.transFigure,
        bbox=dict(boxstyle="larrow,pad=0.1", fc="w", ec="k"),
    )

plt.show()
